import typing
from abc import abstractmethod


class GeneratesSourceCode(typing.Protocol):
    @abstractmethod
    def generate_code(self) -> str:
        """Generates source code to be written to a python module."""
        raise NotImplementedError


class CalculatesImports(typing.Protocol):
    """Interface for an object that can generate the references / imports it needs."""

    @abstractmethod
    def generate_imports(self) -> typing.Set[str]:
        raise NotImplementedError


class Requirable(typing.Protocol):
    """Interface for a type that can emit the imports it needs."""

    @property
    def requires(self) -> typing.Set[str]:
        raise NotImplementedError



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `DOMSnapshot`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/DOMSnapshot/

from __future__ import annotations

import typing
from dataclasses import dataclass

from . import dom
from . import dom_debugger
from . import page


@dataclass
class DOMNode:
    """A Node in the DOM tree."""

    # `Node`'s nodeType. # noqa
    node_type: int
    # `Node`'s nodeName. # noqa
    node_name: str
    # `Node`'s nodeValue. # noqa
    node_value: str
    # `Node`'s id, corresponds to DOM.Node.backendNodeId. # noqa
    backend_node_id: dom.BackendNodeId
    # Only set for textarea elements, contains the text value. # noqa
    text_value: typing.Optional[str]
    # Only set for input elements, contains the input's associated text value. # noqa
    input_value: typing.Optional[str]
    # Only set for radio and checkbox input elements, indicates if the elementhas been checked # noqa
    input_checked: typing.Optional[bool]
    # Only set for option elements, indicates if the element has been selected # noqa
    option_selected: typing.Optional[bool]
    # The indexes of the node's child nodes in the `domNodes` array returned by`getSnapshot`, if any. # noqa
    child_node_indexes: typing.Optional[int]
    # Attributes of an `Element` node. # noqa
    attributes: typing.Optional[NameValue]
    # Indexes of pseudo elements associated with this node in the `domNodes`array returned by `getSnapshot`, if any. # noqa
    pseudo_element_indexes: typing.Optional[int]
    # The index of the node's related layout tree node in the `layoutTreeNodes`array returned by `getSnapshot`, if any. # noqa
    layout_node_index: typing.Optional[int]
    # Document URL that `Document` or `FrameOwner` node points to. # noqa
    document_url: typing.Optional[str]
    # Base URL that `Document` or `FrameOwner` node uses for URL completion. # noqa
    base_url: typing.Optional[str]
    # Only set for documents, contains the document's content language. # noqa
    content_language: typing.Optional[str]
    # Only set for documents, contains the document's character set encoding. # noqa
    document_encoding: typing.Optional[str]
    # `DocumentType` node's publicId. # noqa
    public_id: typing.Optional[str]
    # `DocumentType` node's systemId. # noqa
    system_id: typing.Optional[str]
    # Frame ID for frame owner elements and also for the document node. # noqa
    frame_id: typing.Optional[page.FrameId]
    # The index of a frame owner element's content document in the `domNodes`array returned by `getSnapshot`, if any. # noqa
    content_document_index: typing.Optional[int]
    # Type of a pseudo element node. # noqa
    pseudo_type: typing.Optional[dom.PseudoType]
    # Shadow root type. # noqa
    shadow_root_type: typing.Optional[dom.ShadowRootType]
    # Whether this DOM node responds to mouse clicks. This includes nodes thathave had click event listeners attached via JavaScript as well as anchor tagsthat naturally navigate when clicked. # noqa
    is_clickable: typing.Optional[bool]
    # Details of the node's event listeners, if any. # noqa
    event_listeners: typing.Optional[dom_debugger.EventListener]
    # The selected url for nodes with a srcset attribute. # noqa
    current_source_url: typing.Optional[str]
    # The url of the script (if any) that generates this node. # noqa
    origin_url: typing.Optional[str]
    # Scroll offsets, set when this node is a Document. # noqa
    scroll_offset_x: typing.Optional[float]
    # Description is missing from the devtools protocol document. # noqa
    scroll_offset_y: typing.Optional[float]


@dataclass
class InlineTextBox:
    """Details of post layout rendered text positions.

    The exact layout should not be regarded as stable and may change between versions.
    """

    # The bounding box in document coordinates. Note that scroll offset of thedocument is ignored. # noqa
    bounding_box: dom.Rect
    # The starting index in characters, for this post layout textbox substring.Characters that would be represented as a surrogate pair in UTF-16 have length2. # noqa
    start_character_index: int
    # The number of characters in this post layout textbox substring. Charactersthat would be represented as a surrogate pair in UTF-16 have length 2. # noqa
    num_characters: int


@dataclass
class LayoutTreeNode:
    """Details of an element in the DOM tree with a LayoutObject."""

    # The index of the related DOM node in the `domNodes` array returned by`getSnapshot`. # noqa
    dom_node_index: int
    # The bounding box in document coordinates. Note that scroll offset of thedocument is ignored. # noqa
    bounding_box: dom.Rect
    # Contents of the LayoutText, if any. # noqa
    layout_text: typing.Optional[str]
    # The post-layout inline text nodes, if any. # noqa
    inline_text_nodes: typing.Optional[InlineTextBox]
    # Index into the `computedStyles` array returned by `getSnapshot`. # noqa
    style_index: typing.Optional[int]
    # Global paint order index, which is determined by the stacking order of thenodes. Nodes that are painted together will have the same index. Only providedif includePaintOrder in getSnapshot was true. # noqa
    paint_order: typing.Optional[int]
    # Set to true to indicate the element begins a new stacking context. # noqa
    is_stacking_context: typing.Optional[bool]


@dataclass
class ComputedStyle:
    """A subset of the full ComputedStyle as defined by the request whitelist."""

    # Name/value pairs of computed style properties. # noqa
    properties: NameValue


@dataclass
class NameValue:
    """A name/value pair."""

    # Attribute/property name. # noqa
    name: str
    # Attribute/property value. # noqa
    value: str


class StringIndex(int):
    """Index of the string in the strings table."""

    def to_json(self) -> StringIndex:
        return self

    @classmethod
    def from_json(cls, value: int) -> StringIndex:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


@dataclass
class ArrayOfStrings:
    """Index of the string in the strings table."""


@dataclass
class RareStringData:
    """Data that is only present on rare nodes."""

    # Description is missing from the devtools protocol document. # noqa
    index: int
    # Description is missing from the devtools protocol document. # noqa
    value: StringIndex


@dataclass
class RareBooleanData:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    index: int


@dataclass
class RareIntegerData:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    index: int
    # Description is missing from the devtools protocol document. # noqa
    value: int


@dataclass
class Rectangle:
    """Description is missing from the devtools protocol document."""


@dataclass
class DocumentSnapshot:
    """Document snapshot."""

    # Document URL that `Document` or `FrameOwner` node points to. # noqa
    document_url: StringIndex
    # Document title. # noqa
    title: StringIndex
    # Base URL that `Document` or `FrameOwner` node uses for URL completion. # noqa
    base_url: StringIndex
    # Contains the document's content language. # noqa
    content_language: StringIndex
    # Contains the document's character set encoding. # noqa
    encoding_name: StringIndex
    # `DocumentType` node's publicId. # noqa
    public_id: StringIndex
    # `DocumentType` node's systemId. # noqa
    system_id: StringIndex
    # Frame ID for frame owner elements and also for the document node. # noqa
    frame_id: StringIndex
    # A table with dom nodes. # noqa
    nodes: NodeTreeSnapshot
    # The nodes in the layout tree. # noqa
    layout: LayoutTreeSnapshot
    # The post-layout inline text nodes. # noqa
    text_boxes: TextBoxSnapshot
    # Horizontal scroll offset. # noqa
    scroll_offset_x: typing.Optional[float]
    # Vertical scroll offset. # noqa
    scroll_offset_y: typing.Optional[float]
    # Document content width. # noqa
    content_width: typing.Optional[float]
    # Document content height. # noqa
    content_height: typing.Optional[float]


@dataclass
class NodeTreeSnapshot:
    """Table containing nodes."""

    # Parent node index. # noqa
    parent_index: typing.Optional[int]
    # `Node`'s nodeType. # noqa
    node_type: typing.Optional[int]
    # Type of the shadow root the `Node` is in. String values are equal to the`ShadowRootType` enum. # noqa
    shadow_root_type: typing.Optional[RareStringData]
    # `Node`'s nodeName. # noqa
    node_name: typing.Optional[StringIndex]
    # `Node`'s nodeValue. # noqa
    node_value: typing.Optional[StringIndex]
    # `Node`'s id, corresponds to DOM.Node.backendNodeId. # noqa
    backend_node_id: typing.Optional[dom.BackendNodeId]
    # Attributes of an `Element` node. Flatten name, value pairs. # noqa
    attributes: typing.Optional[ArrayOfStrings]
    # Only set for textarea elements, contains the text value. # noqa
    text_value: typing.Optional[RareStringData]
    # Only set for input elements, contains the input's associated text value. # noqa
    input_value: typing.Optional[RareStringData]
    # Only set for radio and checkbox input elements, indicates if the elementhas been checked # noqa
    input_checked: typing.Optional[RareBooleanData]
    # Only set for option elements, indicates if the element has been selected # noqa
    option_selected: typing.Optional[RareBooleanData]
    # The index of the document in the list of the snapshot documents. # noqa
    content_document_index: typing.Optional[RareIntegerData]
    # Type of a pseudo element node. # noqa
    pseudo_type: typing.Optional[RareStringData]
    # Pseudo element identifier for this node. Only present if there is a validpseudoType. # noqa
    pseudo_identifier: typing.Optional[RareStringData]
    # Whether this DOM node responds to mouse clicks. This includes nodes thathave had click event listeners attached via JavaScript as well as anchor tagsthat naturally navigate when clicked. # noqa
    is_clickable: typing.Optional[RareBooleanData]
    # The selected url for nodes with a srcset attribute. # noqa
    current_source_url: typing.Optional[RareStringData]
    # The url of the script (if any) that generates this node. # noqa
    origin_url: typing.Optional[RareStringData]


@dataclass
class LayoutTreeSnapshot:
    """Table of details of an element in the DOM tree with a LayoutObject."""

    # Index of the corresponding node in the `NodeTreeSnapshot` array returnedby `captureSnapshot`. # noqa
    node_index: int
    # Array of indexes specifying computed style strings, filtered according tothe `computedStyles` parameter passed to `captureSnapshot`. # noqa
    styles: ArrayOfStrings
    # The absolute position bounding box. # noqa
    bounds: Rectangle
    # Contents of the LayoutText, if any. # noqa
    text: StringIndex
    # Stacking context information. # noqa
    stacking_contexts: RareBooleanData
    # Global paint order index, which is determined by the stacking order of thenodes. Nodes that are painted together will have the same index. Only providedif includePaintOrder in captureSnapshot was true. # noqa
    paint_orders: typing.Optional[int]
    # The offset rect of nodes. Only available when includeDOMRects is set totrue # noqa
    offset_rects: typing.Optional[Rectangle]
    # The scroll rect of nodes. Only available when includeDOMRects is set totrue # noqa
    scroll_rects: typing.Optional[Rectangle]
    # The client rect of nodes. Only available when includeDOMRects is set totrue # noqa
    client_rects: typing.Optional[Rectangle]
    # The list of background colors that are blended with colors of overlappingelements. # noqa
    blended_background_colors: typing.Optional[StringIndex]
    # The list of computed text opacities. # noqa
    text_color_opacities: typing.Optional[float]


@dataclass
class TextBoxSnapshot:
    """Table of details of the post layout rendered text positions.

    The exact layout should not be regarded as stable and may change between versions.
    """

    # Index of the layout tree node that owns this box collection. # noqa
    layout_index: int
    # The absolute position bounding box. # noqa
    bounds: Rectangle
    # The starting index in characters, for this post layout textbox substring.Characters that would be represented as a surrogate pair in UTF-16 have length2. # noqa
    start: int
    # The number of characters in this post layout textbox substring. Charactersthat would be represented as a surrogate pair in UTF-16 have length 2. # noqa
    length: int


async def disable() -> None:
    """Disables DOM snapshot agent for the given page.

    # noqa
    """
    ...


async def enable() -> None:
    """Enables DOM snapshot agent for the given page.

    # noqa
    """
    ...


async def get_snapshot() -> None:
    """Returns a document snapshot, including the full DOM tree of the root node (including iframes, template contents,
    and imported documents) in a flattened array, as well as layout and white-listed computed style information for the
    nodes.

    Shadow DOM in the returned DOM tree is flattened. # noqa
    """
    ...


async def capture_snapshot() -> None:
    """Returns a document snapshot, including the full DOM tree of the root node (including iframes, template contents,
    and imported documents) in a flattened array, as well as layout and white-listed computed style information for the
    nodes.

    Shadow DOM in the returned DOM tree is flattened. # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `CSS`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/CSS/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from . import dom
from . import page
from .utils import memoize_event


class StyleSheetId(str):
    """Description is missing from the devtools protocol document."""

    def to_json(self) -> StyleSheetId:
        return self

    @classmethod
    def from_json(cls, value: str) -> StyleSheetId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class StyleSheetOrigin(str, enum.Enum):
    """Stylesheet type: "injected" for stylesheets injected via extension, "user-agent" for user-agent stylesheets,
    "inspector" for stylesheets created by the inspector (i.e. those holding the "via inspector" rules), "regular" for
    regular stylesheets."""

    INJECTED = "injected"
    USER_AGENT = "user-agent"
    INSPECTOR = "inspector"
    REGULAR = "regular"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class PseudoElementMatches:
    """CSS rule collection for a single pseudo style."""

    # Pseudo element type. # noqa
    pseudo_type: dom.PseudoType
    # Matches of CSS rules applicable to the pseudo style. # noqa
    matches: RuleMatch
    # Pseudo element custom ident. # noqa
    pseudo_identifier: typing.Optional[str]


@dataclass
class InheritedStyleEntry:
    """Inherited CSS rule collection from ancestor node."""

    # Matches of CSS rules matching the ancestor node in the style inheritancechain. # noqa
    matched_css_rules: RuleMatch
    # The ancestor node's inline style, if any, in the style inheritance chain. # noqa
    inline_style: typing.Optional[CSSStyle]


@dataclass
class InheritedPseudoElementMatches:
    """Inherited pseudo element matches from pseudos of an ancestor node."""

    # Matches of pseudo styles from the pseudos of an ancestor node. # noqa
    pseudo_elements: PseudoElementMatches


@dataclass
class RuleMatch:
    """Match data for a CSS rule."""

    # CSS rule in the match. # noqa
    rule: CSSRule
    # Matching selector indices in the rule's selectorList selectors (0-based). # noqa
    matching_selectors: int


@dataclass
class Value:
    """Data for a simple selector (these are delimited by commas in a selector list)."""

    # Value text. # noqa
    text: str
    # Value range in the underlying resource (if available). # noqa
    range: typing.Optional[SourceRange]


@dataclass
class SelectorList:
    """Selector list data."""

    # Selectors in the list. # noqa
    selectors: Value
    # Rule selector text. # noqa
    text: str


@dataclass
class CSSStyleSheetHeader:
    """CSS stylesheet metainformation."""

    # The stylesheet identifier. # noqa
    style_sheet_id: StyleSheetId
    # Owner frame identifier. # noqa
    frame_id: page.FrameId
    # Stylesheet resource URL. Empty if this is a constructed stylesheet createdusing new CSSStyleSheet() (but non-empty if this is a constructed sylesheetimported as a CSS module script). # noqa
    source_url: str
    # Stylesheet origin. # noqa
    origin: StyleSheetOrigin
    # Stylesheet title. # noqa
    title: str
    # Denotes whether the stylesheet is disabled. # noqa
    disabled: bool
    # Whether this stylesheet is created for STYLE tag by parser. This flag isnot set for document.written STYLE tags. # noqa
    is_inline: bool
    # Whether this stylesheet is mutable. Inline stylesheets become mutableafter they have been modified via CSSOM API. <link> element's stylesheets becomemutable only if DevTools modifies them. Constructed stylesheets (newCSSStyleSheet()) are mutable immediately after creation. # noqa
    is_mutable: bool
    # True if this stylesheet is created through new CSSStyleSheet() or importedas a CSS module script. # noqa
    is_constructed: bool
    # Line offset of the stylesheet within the resource (zero based). # noqa
    start_line: float
    # Column offset of the stylesheet within the resource (zero based). # noqa
    start_column: float
    # Size of the content (in characters). # noqa
    length: float
    # Line offset of the end of the stylesheet within the resource (zero based). # noqa
    end_line: float
    # Column offset of the end of the stylesheet within the resource (zerobased). # noqa
    end_column: float
    # URL of source map associated with the stylesheet (if any). # noqa
    source_map_url: typing.Optional[str]
    # The backend id for the owner node of the stylesheet. # noqa
    owner_node: typing.Optional[dom.BackendNodeId]
    # Whether the sourceURL field value comes from the sourceURL comment. # noqa
    has_source_url: typing.Optional[bool]


@dataclass
class CSSRule:
    """CSS rule representation."""

    # Rule selector data. # noqa
    selector_list: SelectorList
    # Parent stylesheet's origin. # noqa
    origin: StyleSheetOrigin
    # Associated style declaration. # noqa
    style: CSSStyle
    # The css style sheet identifier (absent for user agent stylesheet and user-specified stylesheet rules) this rule came from. # noqa
    style_sheet_id: typing.Optional[StyleSheetId]
    # Array of selectors from ancestor style rules, sorted by distance from thecurrent rule. # noqa
    nesting_selectors: typing.Optional[str]
    # Media list array (for rules involving media queries). The array enumeratesmedia queries starting with the innermost one, going outwards. # noqa
    media: typing.Optional[CSSMedia]
    # Container query list array (for rules involving container queries). Thearray enumerates container queries starting with the innermost one, goingoutwards. # noqa
    container_queries: typing.Optional[CSSContainerQuery]
    # @supports CSS at-rule array. The array enumerates @supports at-rulesstarting with the innermost one, going outwards. # noqa
    supports: typing.Optional[CSSSupports]
    # Cascade layer array. Contains the layer hierarchy that this rule belongsto starting with the innermost layer and going outwards. # noqa
    layers: typing.Optional[CSSLayer]
    # @scope CSS at-rule array. The array enumerates @scope at-rules startingwith the innermost one, going outwards. # noqa
    scopes: typing.Optional[CSSScope]


@dataclass
class RuleUsage:
    """CSS coverage information."""

    # The css style sheet identifier (absent for user agent stylesheet and user-specified stylesheet rules) this rule came from. # noqa
    style_sheet_id: StyleSheetId
    # Offset of the start of the rule (including selector) from the beginning ofthe stylesheet. # noqa
    start_offset: float
    # Offset of the end of the rule body from the beginning of the stylesheet. # noqa
    end_offset: float
    # Indicates whether the rule was actually used by some element in the page. # noqa
    used: bool


@dataclass
class SourceRange:
    """Text range within a resource.

    All numbers are zero-based.
    """

    # Start line of range. # noqa
    start_line: int
    # Start column of range (inclusive). # noqa
    start_column: int
    # End line of range # noqa
    end_line: int
    # End column of range (exclusive). # noqa
    end_column: int


@dataclass
class ShorthandEntry:
    """Description is missing from the devtools protocol document."""

    # Shorthand name. # noqa
    name: str
    # Shorthand value. # noqa
    value: str
    # Whether the property has "!important" annotation (implies `false` ifabsent). # noqa
    important: typing.Optional[bool]


@dataclass
class CSSComputedStyleProperty:
    """Description is missing from the devtools protocol document."""

    # Computed style property name. # noqa
    name: str
    # Computed style property value. # noqa
    value: str


@dataclass
class CSSStyle:
    """CSS style representation."""

    # CSS properties in the style. # noqa
    css_properties: CSSProperty
    # Computed values for all shorthands found in the style. # noqa
    shorthand_entries: ShorthandEntry
    # The css style sheet identifier (absent for user agent stylesheet and user-specified stylesheet rules) this rule came from. # noqa
    style_sheet_id: typing.Optional[StyleSheetId]
    # Style declaration text (if available). # noqa
    css_text: typing.Optional[str]
    # Style declaration range in the enclosing stylesheet (if available). # noqa
    range: typing.Optional[SourceRange]


@dataclass
class CSSProperty:
    """CSS property declaration data."""

    # The property name. # noqa
    name: str
    # The property value. # noqa
    value: str
    # Whether the property has "!important" annotation (implies `false` ifabsent). # noqa
    important: typing.Optional[bool]
    # Whether the property is implicit (implies `false` if absent). # noqa
    implicit: typing.Optional[bool]
    # The full property text as specified in the style. # noqa
    text: typing.Optional[str]
    # Whether the property is understood by the browser (implies `true` ifabsent). # noqa
    parsed_ok: typing.Optional[bool]
    # Whether the property is disabled by the user (present for source-basedproperties only). # noqa
    disabled: typing.Optional[bool]
    # The entire property range in the enclosing style declaration (ifavailable). # noqa
    range: typing.Optional[SourceRange]
    # Parsed longhand components of this property if it is a shorthand. Thisfield will be empty if the given property is not a shorthand. # noqa
    longhand_properties: typing.Optional[CSSProperty]


@dataclass
class CSSMedia:
    """CSS media rule descriptor."""

    # Media query text. # noqa
    text: str
    # Source of the media query: "mediaRule" if specified by a @media rule,"importRule" if specified by an @import rule, "linkedSheet" if specified by a"media" attribute in a linked stylesheet's LINK tag, "inlineSheet" if specifiedby a "media" attribute in an inline stylesheet's STYLE tag. # noqa
    source: typing.List[typing.Literal["mediaRule", "importRule", "linkedSheet", "inlineSheet"]]
    # URL of the document containing the media query description. # noqa
    source_url: typing.Optional[str]
    # The associated rule (@media or @import) header range in the enclosingstylesheet (if available). # noqa
    range: typing.Optional[SourceRange]
    # Identifier of the stylesheet containing this object (if exists). # noqa
    style_sheet_id: typing.Optional[StyleSheetId]
    # Array of media queries. # noqa
    media_list: typing.Optional[MediaQuery]


@dataclass
class MediaQuery:
    """Media query descriptor."""

    # Array of media query expressions. # noqa
    expressions: MediaQueryExpression
    # Whether the media query condition is satisfied. # noqa
    active: bool


@dataclass
class MediaQueryExpression:
    """Media query expression descriptor."""

    # Media query expression value. # noqa
    value: float
    # Media query expression units. # noqa
    unit: str
    # Media query expression feature. # noqa
    feature: str
    # The associated range of the value text in the enclosing stylesheet (ifavailable). # noqa
    value_range: typing.Optional[SourceRange]
    # Computed length of media query expression (if applicable). # noqa
    computed_length: typing.Optional[float]


@dataclass
class CSSContainerQuery:
    """CSS container query rule descriptor."""

    # Container query text. # noqa
    text: str
    # The associated rule header range in the enclosing stylesheet (ifavailable). # noqa
    range: typing.Optional[SourceRange]
    # Identifier of the stylesheet containing this object (if exists). # noqa
    style_sheet_id: typing.Optional[StyleSheetId]
    # Optional name for the container. # noqa
    name: typing.Optional[str]
    # Optional physical axes queried for the container. # noqa
    physical_axes: typing.Optional[dom.PhysicalAxes]
    # Optional logical axes queried for the container. # noqa
    logical_axes: typing.Optional[dom.LogicalAxes]


@dataclass
class CSSSupports:
    """CSS Supports at-rule descriptor."""

    # Supports rule text. # noqa
    text: str
    # Whether the supports condition is satisfied. # noqa
    active: bool
    # The associated rule header range in the enclosing stylesheet (ifavailable). # noqa
    range: typing.Optional[SourceRange]
    # Identifier of the stylesheet containing this object (if exists). # noqa
    style_sheet_id: typing.Optional[StyleSheetId]


@dataclass
class CSSScope:
    """CSS Scope at-rule descriptor."""

    # Scope rule text. # noqa
    text: str
    # The associated rule header range in the enclosing stylesheet (ifavailable). # noqa
    range: typing.Optional[SourceRange]
    # Identifier of the stylesheet containing this object (if exists). # noqa
    style_sheet_id: typing.Optional[StyleSheetId]


@dataclass
class CSSLayer:
    """CSS Layer at-rule descriptor."""

    # Layer name. # noqa
    text: str
    # The associated rule header range in the enclosing stylesheet (ifavailable). # noqa
    range: typing.Optional[SourceRange]
    # Identifier of the stylesheet containing this object (if exists). # noqa
    style_sheet_id: typing.Optional[StyleSheetId]


@dataclass
class CSSLayerData:
    """CSS Layer data."""

    # Layer name. # noqa
    name: str
    # Layer order. The order determines the order of the layer in the cascadeorder. A higher number has higher priority in the cascade order. # noqa
    order: float
    # Direct sub-layers # noqa
    sub_layers: typing.Optional[CSSLayerData]


@dataclass
class PlatformFontUsage:
    """Information about amount of glyphs that were rendered with given font."""

    # Font's family name reported by platform. # noqa
    family_name: str
    # Indicates if the font was downloaded or resolved locally. # noqa
    is_custom_font: bool
    # Amount of glyphs that were rendered with this font. # noqa
    glyph_count: float


@dataclass
class FontVariationAxis:
    """Information about font variation axes for variable fonts."""

    # The font-variation-setting tag (a.k.a. "axis tag"). # noqa
    tag: str
    # Human-readable variation name in the default language (normally, "en"). # noqa
    name: str
    # The minimum value (inclusive) the font supports for this tag. # noqa
    min_value: float
    # The maximum value (inclusive) the font supports for this tag. # noqa
    max_value: float
    # The default value. # noqa
    default_value: float


@dataclass
class FontFace:
    """Properties of a web font: https://www.w3.org/TR/2008/REC-CSS2-20080411/fonts.html#font-descriptions and
    additional information such as platformFontFamily and fontVariationAxes."""

    # The font-family. # noqa
    font_family: str
    # The font-style. # noqa
    font_style: str
    # The font-variant. # noqa
    font_variant: str
    # The font-weight. # noqa
    font_weight: str
    # The font-stretch. # noqa
    font_stretch: str
    # The font-display. # noqa
    font_display: str
    # The unicode-range. # noqa
    unicode_range: str
    # The src. # noqa
    src: str
    # The resolved platform font family # noqa
    platform_font_family: str
    # Available variation settings (a.k.a. "axes"). # noqa
    font_variation_axes: typing.Optional[FontVariationAxis]


@dataclass
class CSSTryRule:
    """CSS try rule representation."""

    # Parent stylesheet's origin. # noqa
    origin: StyleSheetOrigin
    # The css style sheet identifier (absent for user agent stylesheet and user-specified stylesheet rules) this rule came from. # noqa
    style_sheet_id: typing.Optional[StyleSheetId]
    # Associated style declaration. # noqa
    style: typing.Optional[CSSStyle]


@dataclass
class CSSPositionFallbackRule:
    """CSS position-fallback rule representation."""

    # Description is missing from the devtools protocol document. # noqa
    name: Value
    # List of keyframes. # noqa
    try_rules: CSSTryRule


@dataclass
class CSSKeyframesRule:
    """CSS keyframes rule representation."""

    # Animation name. # noqa
    animation_name: Value
    # List of keyframes. # noqa
    keyframes: CSSKeyframeRule


@dataclass
class CSSKeyframeRule:
    """CSS keyframe rule representation."""

    # Parent stylesheet's origin. # noqa
    origin: StyleSheetOrigin
    # Associated key text. # noqa
    key_text: Value
    # Associated style declaration. # noqa
    style: CSSStyle
    # The css style sheet identifier (absent for user agent stylesheet and user-specified stylesheet rules) this rule came from. # noqa
    style_sheet_id: typing.Optional[StyleSheetId]


@dataclass
class StyleDeclarationEdit:
    """A descriptor of operation to mutate style declaration text."""

    # The css style sheet identifier. # noqa
    style_sheet_id: StyleSheetId
    # The range of the style text in the enclosing stylesheet. # noqa
    range: SourceRange
    # New style text. # noqa
    text: str


@dataclass
@memoize_event("CSS.fontsUpdated")
class FontsUpdated:
    """Fires whenever a web font is updated.

    A non-empty font parameter indicates a successfully loaded web font.
    """

    font: typing.Optional[FontFace]


@dataclass
@memoize_event("CSS.mediaQueryResultChanged")
class MediaQueryResultChanged:
    """Fires whenever a MediaQuery result changes (for example, after a browser window has been resized.) The current
    implementation considers only viewport-dependent media features."""


@dataclass
@memoize_event("CSS.styleSheetAdded")
class StyleSheetAdded:
    """Fired whenever an active document stylesheet is added."""

    header: CSSStyleSheetHeader


@dataclass
@memoize_event("CSS.styleSheetChanged")
class StyleSheetChanged:
    """Fired whenever a stylesheet is changed as a result of the client operation."""

    style_sheet_id: StyleSheetId


@dataclass
@memoize_event("CSS.styleSheetRemoved")
class StyleSheetRemoved:
    """Fired whenever an active document stylesheet is removed."""

    style_sheet_id: StyleSheetId


async def add_rule() -> None:
    """Inserts a new rule with the given `ruleText` in a stylesheet with given `styleSheetId`, at the position specified
    by `location`.

    # noqa
    """
    ...


async def collect_class_names() -> None:
    """Returns all class names from specified stylesheet.

    # noqa
    """
    ...


async def create_style_sheet() -> None:
    """Creates a new special "via-inspector" stylesheet in the frame with given `frameId`.

    # noqa
    """
    ...


async def disable() -> None:
    """Disables the CSS agent for the given page.

    # noqa
    """
    ...


async def enable() -> None:
    """Enables the CSS agent for the given page.

    Clients should not assume that the CSS agent has been enabled until the result of this command is received. # noqa
    """
    ...


async def force_pseudo_state() -> None:
    """Ensures that the given node will have specified pseudo-classes whenever its style is computed by the browser.

    # noqa
    """
    ...


async def get_background_colors() -> None:
    """Description is missing from the devtools protocol document.

    # noqa
    """
    ...


async def get_computed_style_for_node() -> None:
    """Returns the computed style for a DOM node identified by `nodeId`.

    # noqa
    """
    ...


async def get_inline_styles_for_node() -> None:
    """Returns the styles defined inline (explicitly in the "style" attribute and implicitly, using DOM attributes) for
    a DOM node identified by `nodeId`.

    # noqa
    """
    ...


async def get_matched_styles_for_node() -> None:
    """Returns requested styles for a DOM node identified by `nodeId`.

    # noqa
    """
    ...


async def get_media_queries() -> None:
    """Returns all media queries parsed by the rendering engine.

    # noqa
    """
    ...


async def get_platform_fonts_for_node() -> None:
    """Requests information about platform fonts which we used to render child TextNodes in the given node.

    # noqa
    """
    ...


async def get_style_sheet_text() -> None:
    """Returns the current textual content for a stylesheet.

    # noqa
    """
    ...


async def get_layers_for_node() -> None:
    """Returns all layers parsed by the rendering engine for the tree scope of a node.

    Given a DOM element identified by nodeId, getLayersForNode returns the root layer for the nearest ancestor document
    or shadow root. The layer root contains the full layer tree for the tree scope and their ordering. # noqa
    """
    ...


async def track_computed_style_updates() -> None:
    """Starts tracking the given computed styles for updates.

    The specified array of properties replaces the one previously specified. Pass empty array to disable tracking. Use
    takeComputedStyleUpdates to retrieve the list of nodes that had properties modified. The changes to computed style
    properties are only tracked for nodes pushed to the front-end by the DOM agent. If no changes to the tracked
    properties occur after the node has been pushed to the front-end, no updates will be issued for the node. # noqa
    """
    ...


async def take_computed_style_updates() -> None:
    """Polls the next batch of computed style updates.

    # noqa
    """
    ...


async def set_effective_property_value_for_node() -> None:
    """Find a rule with the given active property for the given node and set the new value for this property # noqa."""
    ...


async def set_keyframe_key() -> None:
    """Modifies the keyframe rule key text.

    # noqa
    """
    ...


async def set_media_text() -> None:
    """Modifies the rule selector.

    # noqa
    """
    ...


async def set_container_query_text() -> None:
    """Modifies the expression of a container query.

    # noqa
    """
    ...


async def set_supports_text() -> None:
    """Modifies the expression of a supports at-rule.

    # noqa
    """
    ...


async def set_scope_text() -> None:
    """Modifies the expression of a scope at-rule.

    # noqa
    """
    ...


async def set_rule_selector() -> None:
    """Modifies the rule selector.

    # noqa
    """
    ...


async def set_style_sheet_text() -> None:
    """Sets the new stylesheet text.

    # noqa
    """
    ...


async def set_style_texts() -> None:
    """Applies specified style edits one after another in the given order.

    # noqa
    """
    ...


async def start_rule_usage_tracking() -> None:
    """Enables the selector recording.

    # noqa
    """
    ...


async def stop_rule_usage_tracking() -> None:
    """Stop tracking rule usage and return the list of rules that were used since last call to `takeCoverageDelta` (or
    since start of coverage instrumentation).

    # noqa
    """
    ...


async def take_coverage_delta() -> None:
    """Obtain list of rules that became used since last call to this method (or since start of coverage
    instrumentation).

    # noqa
    """
    ...


async def set_local_fonts_enabled() -> None:
    """Enables/disables rendering of local CSS fonts (enabled by default).

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `Audits`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/Audits/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from . import dom
from . import network
from . import page
from . import runtime
from .utils import memoize_event


@dataclass
class AffectedCookie:
    """Information about a cookie that is affected by an inspector issue."""

    # The following three properties uniquely identify a cookie # noqa
    name: str
    # Description is missing from the devtools protocol document. # noqa
    path: str
    # Description is missing from the devtools protocol document. # noqa
    domain: str


@dataclass
class AffectedRequest:
    """Information about a request that is affected by an inspector issue."""

    # The unique request id. # noqa
    request_id: network.RequestId
    # Description is missing from the devtools protocol document. # noqa
    url: typing.Optional[str]


@dataclass
class AffectedFrame:
    """Information about the frame affected by an inspector issue."""

    # Description is missing from the devtools protocol document. # noqa
    frame_id: page.FrameId


class CookieExclusionReason(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    EXCLUDE_SAME_SITE_UNSPECIFIED_TREATED_AS_LAX = "exclude_same_site_unspecified_treated_as_lax"
    EXCLUDE_SAME_SITE_NONE_INSECURE = "exclude_same_site_none_insecure"
    EXCLUDE_SAME_SITE_LAX = "exclude_same_site_lax"
    EXCLUDE_SAME_SITE_STRICT = "exclude_same_site_strict"
    EXCLUDE_INVALID_SAME_PARTY = "exclude_invalid_same_party"
    EXCLUDE_SAME_PARTY_CROSS_PARTY_CONTEXT = "exclude_same_party_cross_party_context"
    EXCLUDE_DOMAIN_NON_A_S_C_I_I = "exclude_domain_non_ascii"
    EXCLUDE_THIRD_PARTY_COOKIE_BLOCKED_IN_FIRST_PARTY_SET = "exclude_third_party_cookie_blocked_in_first_party_set"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class CookieWarningReason(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    WARN_SAME_SITE_UNSPECIFIED_CROSS_SITE_CONTEXT = "warn_same_site_unspecified_cross_site_context"
    WARN_SAME_SITE_NONE_INSECURE = "warn_same_site_none_insecure"
    WARN_SAME_SITE_UNSPECIFIED_LAX_ALLOW_UNSAFE = "warn_same_site_unspecified_lax_allow_unsafe"
    WARN_SAME_SITE_STRICT_LAX_DOWNGRADE_STRICT = "warn_same_site_strict_lax_downgrade_strict"
    WARN_SAME_SITE_STRICT_CROSS_DOWNGRADE_STRICT = "warn_same_site_strict_cross_downgrade_strict"
    WARN_SAME_SITE_STRICT_CROSS_DOWNGRADE_LAX = "warn_same_site_strict_cross_downgrade_lax"
    WARN_SAME_SITE_LAX_CROSS_DOWNGRADE_STRICT = "warn_same_site_lax_cross_downgrade_strict"
    WARN_SAME_SITE_LAX_CROSS_DOWNGRADE_LAX = "warn_same_site_lax_cross_downgrade_lax"
    WARN_ATTRIBUTE_VALUE_EXCEEDS_MAX_SIZE = "warn_attribute_value_exceeds_max_size"
    WARN_DOMAIN_NON_A_S_C_I_I = "warn_domain_non_ascii"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class CookieOperation(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    SET_COOKIE = "set_cookie"
    READ_COOKIE = "read_cookie"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class CookieIssueDetails:
    """This information is currently necessary, as the front-end has a difficult time finding a specific cookie.

    With this, we can convey specific error information without the cookie.
    """

    # Description is missing from the devtools protocol document. # noqa
    cookie_warning_reasons: CookieWarningReason
    # Description is missing from the devtools protocol document. # noqa
    cookie_exclusion_reasons: CookieExclusionReason
    # Optionally identifies the site-for-cookies and the cookie url, which maybe used by the front-end as additional context. # noqa
    operation: CookieOperation
    # If AffectedCookie is not set then rawCookieLine contains the raw Set-Cookie header string. This hints at a problem where the cookie line issyntactically or semantically malformed in a way that no valid cookie could becreated. # noqa
    cookie: typing.Optional[AffectedCookie]
    # Description is missing from the devtools protocol document. # noqa
    raw_cookie_line: typing.Optional[str]
    # Description is missing from the devtools protocol document. # noqa
    site_for_cookies: typing.Optional[str]
    # Description is missing from the devtools protocol document. # noqa
    cookie_url: typing.Optional[str]
    # Description is missing from the devtools protocol document. # noqa
    request: typing.Optional[AffectedRequest]


class MixedContentResolutionStatus(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    MIXED_CONTENT_BLOCKED = "mixed_content_blocked"
    MIXED_CONTENT_AUTOMATICALLY_UPGRADED = "mixed_content_automatically_upgraded"
    MIXED_CONTENT_WARNING = "mixed_content_warning"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class MixedContentResourceType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    ATTRIBUTION_SRC = "attribution_src"
    AUDIO = "audio"
    BEACON = "beacon"
    C_S_P_REPORT = "csp_report"
    DOWNLOAD = "download"
    EVENT_SOURCE = "event_source"
    FAVICON = "favicon"
    FONT = "font"
    FORM = "form"
    FRAME = "frame"
    IMAGE = "image"
    IMPORT = "import"
    MANIFEST = "manifest"
    PING = "ping"
    PLUGIN_DATA = "plugin_data"
    PLUGIN_RESOURCE = "plugin_resource"
    PREFETCH = "prefetch"
    RESOURCE = "resource"
    SCRIPT = "script"
    SERVICE_WORKER = "service_worker"
    SHARED_WORKER = "shared_worker"
    STYLESHEET = "stylesheet"
    TRACK = "track"
    VIDEO = "video"
    WORKER = "worker"
    X_M_L_HTTP_REQUEST = "xml_http_request"
    X_S_L_T = "xslt"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class MixedContentIssueDetails:
    """Description is missing from the devtools protocol document."""

    # The way the mixed content issue is being resolved. # noqa
    resolution_status: MixedContentResolutionStatus
    # The unsafe http url causing the mixed content issue. # noqa
    insecure_url: str
    # The url responsible for the call to an unsafe url. # noqa
    main_resource_url: str
    # The type of resource causing the mixed content issue (css, js, iframe,form,...). Marked as optional because it is mapped to fromblink::mojom::RequestContextType, which will be replaced bynetwork::mojom::RequestDestination # noqa
    resource_type: typing.Optional[MixedContentResourceType]
    # The mixed content request. Does not always exist (e.g. for unsafe formsubmission urls). # noqa
    request: typing.Optional[AffectedRequest]
    # Optional because not every mixed content issue is necessarily linked to aframe. # noqa
    frame: typing.Optional[AffectedFrame]


class BlockedByResponseReason(str, enum.Enum):
    """Enum indicating the reason a response has been blocked.

    These reasons are refinements of the net error BLOCKED_BY_RESPONSE.
    """

    COEP_FRAME_RESOURCE_NEEDS_COEP_HEADER = "coep_frame_resource_needs_coep_header"
    COOP_SANDBOXED_I_FRAME_CANNOT_NAVIGATE_TO_COOP_PAGE = "coop_sandboxed_i_frame_cannot_navigate_to_coop_page"
    CORP_NOT_SAME_ORIGIN = "corp_not_same_origin"
    CORP_NOT_SAME_ORIGIN_AFTER_DEFAULTED_TO_SAME_ORIGIN_BY_COEP = (
        "corp_not_same_origin_after_defaulted_to_same_origin_by_coep"
    )
    CORP_NOT_SAME_SITE = "corp_not_same_site"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class BlockedByResponseIssueDetails:
    """Details for a request that has been blocked with the BLOCKED_BY_RESPONSE code.

    Currently only used for COEP/COOP, but may be extended to include some CSP errors in the future.
    """

    # Description is missing from the devtools protocol document. # noqa
    request: AffectedRequest
    # Description is missing from the devtools protocol document. # noqa
    reason: BlockedByResponseReason
    # Description is missing from the devtools protocol document. # noqa
    parent_frame: typing.Optional[AffectedFrame]
    # Description is missing from the devtools protocol document. # noqa
    blocked_frame: typing.Optional[AffectedFrame]


class HeavyAdResolutionStatus(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    HEAVY_AD_BLOCKED = "heavy_ad_blocked"
    HEAVY_AD_WARNING = "heavy_ad_warning"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class HeavyAdReason(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    NETWORK_TOTAL_LIMIT = "network_total_limit"
    CPU_TOTAL_LIMIT = "cpu_total_limit"
    CPU_PEAK_LIMIT = "cpu_peak_limit"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class HeavyAdIssueDetails:
    """Description is missing from the devtools protocol document."""

    # The resolution status, either blocking the content or warning. # noqa
    resolution: HeavyAdResolutionStatus
    # The reason the ad was blocked, total network or cpu or peak cpu. # noqa
    reason: HeavyAdReason
    # The frame that was blocked. # noqa
    frame: AffectedFrame


class ContentSecurityPolicyViolationType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    K_INLINE_VIOLATION = "k_inline_violation"
    K_EVAL_VIOLATION = "k_eval_violation"
    K_U_R_L_VIOLATION = "k_url_violation"
    K_TRUSTED_TYPES_SINK_VIOLATION = "k_trusted_types_sink_violation"
    K_TRUSTED_TYPES_POLICY_VIOLATION = "k_trusted_types_policy_violation"
    K_WASM_EVAL_VIOLATION = "k_wasm_eval_violation"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class SourceCodeLocation:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    url: str
    # Description is missing from the devtools protocol document. # noqa
    line_number: int
    # Description is missing from the devtools protocol document. # noqa
    column_number: int
    # Description is missing from the devtools protocol document. # noqa
    script_id: typing.Optional[runtime.ScriptId]


@dataclass
class ContentSecurityPolicyIssueDetails:
    """Description is missing from the devtools protocol document."""

    # Specific directive that is violated, causing the CSP issue. # noqa
    violated_directive: str
    # Description is missing from the devtools protocol document. # noqa
    is_report_only: bool
    # Description is missing from the devtools protocol document. # noqa
    content_security_policy_violation_type: ContentSecurityPolicyViolationType
    # The url not included in allowed sources. # noqa
    blocked_url: typing.Optional[str]
    # Description is missing from the devtools protocol document. # noqa
    frame_ancestor: typing.Optional[AffectedFrame]
    # Description is missing from the devtools protocol document. # noqa
    source_code_location: typing.Optional[SourceCodeLocation]
    # Description is missing from the devtools protocol document. # noqa
    violating_node_id: typing.Optional[dom.BackendNodeId]


class SharedArrayBufferIssueType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    TRANSFER_ISSUE = "transfer_issue"
    CREATION_ISSUE = "creation_issue"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class SharedArrayBufferIssueDetails:
    """Details for a issue arising from an SAB being instantiated in, or transferred to a context that is not cross-
    origin isolated."""

    # Description is missing from the devtools protocol document. # noqa
    source_code_location: SourceCodeLocation
    # Description is missing from the devtools protocol document. # noqa
    is_warning: bool
    # Description is missing from the devtools protocol document. # noqa
    type: SharedArrayBufferIssueType


class TwaQualityEnforcementViolationType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    K_HTTP_ERROR = "k_http_error"
    K_UNAVAILABLE_OFFLINE = "k_unavailable_offline"
    K_DIGITAL_ASSET_LINKS = "k_digital_asset_links"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class TrustedWebActivityIssueDetails:
    """Description is missing from the devtools protocol document."""

    # The url that triggers the violation. # noqa
    url: str
    # Description is missing from the devtools protocol document. # noqa
    violation_type: TwaQualityEnforcementViolationType
    # Description is missing from the devtools protocol document. # noqa
    http_status_code: typing.Optional[int]
    # The package name of the Trusted Web Activity client app. This field isonly used when violation type is kDigitalAssetLinks. # noqa
    package_name: typing.Optional[str]
    # The signature of the Trusted Web Activity client app. This field is onlyused when violation type is kDigitalAssetLinks. # noqa
    signature: typing.Optional[str]


@dataclass
class LowTextContrastIssueDetails:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    violating_node_id: dom.BackendNodeId
    # Description is missing from the devtools protocol document. # noqa
    violating_node_selector: str
    # Description is missing from the devtools protocol document. # noqa
    contrast_ratio: float
    # Description is missing from the devtools protocol document. # noqa
    threshold_aa: float
    # Description is missing from the devtools protocol document. # noqa
    threshold_aaa: float
    # Description is missing from the devtools protocol document. # noqa
    font_size: str
    # Description is missing from the devtools protocol document. # noqa
    font_weight: str


@dataclass
class CorsIssueDetails:
    """Details for a CORS related issue, e.g. a warning or error related to CORS RFC1918 enforcement."""

    # Description is missing from the devtools protocol document. # noqa
    cors_error_status: network.CorsErrorStatus
    # Description is missing from the devtools protocol document. # noqa
    is_warning: bool
    # Description is missing from the devtools protocol document. # noqa
    request: AffectedRequest
    # Description is missing from the devtools protocol document. # noqa
    location: typing.Optional[SourceCodeLocation]
    # Description is missing from the devtools protocol document. # noqa
    initiator_origin: typing.Optional[str]
    # Description is missing from the devtools protocol document. # noqa
    resource_ip_address_space: typing.Optional[network.IPAddressSpace]
    # Description is missing from the devtools protocol document. # noqa
    client_security_state: typing.Optional[network.ClientSecurityState]


class AttributionReportingIssueType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    PERMISSION_POLICY_DISABLED = "permission_policy_disabled"
    UNTRUSTWORTHY_REPORTING_ORIGIN = "untrustworthy_reporting_origin"
    INSECURE_CONTEXT = "insecure_context"
    INVALID_HEADER = "invalid_header"
    INVALID_REGISTER_TRIGGER_HEADER = "invalid_register_trigger_header"
    INVALID_ELIGIBLE_HEADER = "invalid_eligible_header"
    TOO_MANY_CONCURRENT_REQUESTS = "too_many_concurrent_requests"
    SOURCE_AND_TRIGGER_HEADERS = "source_and_trigger_headers"
    SOURCE_IGNORED = "source_ignored"
    TRIGGER_IGNORED = "trigger_ignored"
    OS_SOURCE_IGNORED = "os_source_ignored"
    OS_TRIGGER_IGNORED = "os_trigger_ignored"
    INVALID_REGISTER_OS_SOURCE_HEADER = "invalid_register_os_source_header"
    INVALID_REGISTER_OS_TRIGGER_HEADER = "invalid_register_os_trigger_header"
    WEB_AND_OS_HEADERS = "web_and_os_headers"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class AttributionReportingIssueDetails:
    """Details for issues around "Attribution Reporting API" usage.

    Explainer: https://github.com/WICG/attribution-reporting-api
    """

    # Description is missing from the devtools protocol document. # noqa
    violation_type: AttributionReportingIssueType
    # Description is missing from the devtools protocol document. # noqa
    request: typing.Optional[AffectedRequest]
    # Description is missing from the devtools protocol document. # noqa
    violating_node_id: typing.Optional[dom.BackendNodeId]
    # Description is missing from the devtools protocol document. # noqa
    invalid_parameter: typing.Optional[str]


@dataclass
class QuirksModeIssueDetails:
    """Details for issues about documents in Quirks Mode or Limited Quirks Mode that affects page layouting."""

    # If false, it means the document's mode is "quirks" instead of "limited-quirks". # noqa
    is_limited_quirks_mode: bool
    # Description is missing from the devtools protocol document. # noqa
    document_node_id: dom.BackendNodeId
    # Description is missing from the devtools protocol document. # noqa
    url: str
    # Description is missing from the devtools protocol document. # noqa
    frame_id: page.FrameId
    # Description is missing from the devtools protocol document. # noqa
    loader_id: network.LoaderId


@dataclass
class NavigatorUserAgentIssueDetails:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    url: str
    # Description is missing from the devtools protocol document. # noqa
    location: typing.Optional[SourceCodeLocation]


class GenericIssueErrorType(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    CROSS_ORIGIN_PORTAL_POST_MESSAGE_ERROR = "cross_origin_portal_post_message_error"
    FORM_LABEL_FOR_NAME_ERROR = "form_label_for_name_error"
    FORM_DUPLICATE_ID_FOR_INPUT_ERROR = "form_duplicate_id_for_input_error"
    FORM_INPUT_WITH_NO_LABEL_ERROR = "form_input_with_no_label_error"
    FORM_AUTOCOMPLETE_ATTRIBUTE_EMPTY_ERROR = "form_autocomplete_attribute_empty_error"
    FORM_EMPTY_ID_AND_NAME_ATTRIBUTES_FOR_INPUT_ERROR = "form_empty_id_and_name_attributes_for_input_error"
    FORM_ARIA_LABELLED_BY_TO_NON_EXISTING_ID = "form_aria_labelled_by_to_non_existing_id"
    FORM_INPUT_ASSIGNED_AUTOCOMPLETE_VALUE_TO_ID_OR_NAME_ATTRIBUTE_ERROR = (
        "form_input_assigned_autocomplete_value_to_id_or_name_attribute_error"
    )
    FORM_LABEL_HAS_NEITHER_FOR_NOR_NESTED_INPUT = "form_label_has_neither_for_nor_nested_input"
    FORM_LABEL_FOR_MATCHES_NON_EXISTING_ID_ERROR = "form_label_for_matches_non_existing_id_error"
    FORM_INPUT_HAS_WRONG_BUT_WELL_INTENDED_AUTOCOMPLETE_VALUE_ERROR = (
        "form_input_has_wrong_but_well_intended_autocomplete_value_error"
    )

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class GenericIssueDetails:
    """Depending on the concrete errorType, different properties are set."""

    # Issues with the same errorType are aggregated in the frontend. # noqa
    error_type: GenericIssueErrorType
    # Description is missing from the devtools protocol document. # noqa
    frame_id: typing.Optional[page.FrameId]
    # Description is missing from the devtools protocol document. # noqa
    violating_node_id: typing.Optional[dom.BackendNodeId]
    # Description is missing from the devtools protocol document. # noqa
    violating_node_attribute: typing.Optional[str]


@dataclass
class DeprecationIssueDetails:
    """This issue tracks information needed to print a deprecation message.

    https://source.chromium.org/chromium/chromium/src/+/main:third_party/blink/renderer/core/frame/third_party/blink/renderer/core/frame/deprecation/README.md
    """

    # Description is missing from the devtools protocol document. # noqa
    source_code_location: SourceCodeLocation
    # One of the deprecation names fromthird_party/blink/renderer/core/frame/deprecation/deprecation.json5 # noqa
    type: str
    # Description is missing from the devtools protocol document. # noqa
    affected_frame: typing.Optional[AffectedFrame]


class ClientHintIssueReason(str, enum.Enum):
    """Description is missing from the devtools protocol document."""

    META_TAG_ALLOW_LIST_INVALID_ORIGIN = "meta_tag_allow_list_invalid_origin"
    META_TAG_MODIFIED_H_T_M_L = "meta_tag_modified_html"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class FederatedAuthRequestIssueDetails:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    federated_auth_request_issue_reason: FederatedAuthRequestIssueReason


class FederatedAuthRequestIssueReason(str, enum.Enum):
    """Represents the failure reason when a federated authentication reason fails.

    Should be updated alongside RequestIdTokenStatus in
    third_party/blink/public/mojom/devtools/inspector_issue.mojom to include
    all cases except for success.
    """

    SHOULD_EMBARGO = "should_embargo"
    TOO_MANY_REQUESTS = "too_many_requests"
    WELL_KNOWN_HTTP_NOT_FOUND = "well_known_http_not_found"
    WELL_KNOWN_NO_RESPONSE = "well_known_no_response"
    WELL_KNOWN_INVALID_RESPONSE = "well_known_invalid_response"
    WELL_KNOWN_LIST_EMPTY = "well_known_list_empty"
    CONFIG_NOT_IN_WELL_KNOWN = "config_not_in_well_known"
    WELL_KNOWN_TOO_BIG = "well_known_too_big"
    CONFIG_HTTP_NOT_FOUND = "config_http_not_found"
    CONFIG_NO_RESPONSE = "config_no_response"
    CONFIG_INVALID_RESPONSE = "config_invalid_response"
    CLIENT_METADATA_HTTP_NOT_FOUND = "client_metadata_http_not_found"
    CLIENT_METADATA_NO_RESPONSE = "client_metadata_no_response"
    CLIENT_METADATA_INVALID_RESPONSE = "client_metadata_invalid_response"
    DISABLED_IN_SETTINGS = "disabled_in_settings"
    ERROR_FETCHING_SIGNIN = "error_fetching_signin"
    INVALID_SIGNIN_RESPONSE = "invalid_signin_response"
    ACCOUNTS_HTTP_NOT_FOUND = "accounts_http_not_found"
    ACCOUNTS_NO_RESPONSE = "accounts_no_response"
    ACCOUNTS_INVALID_RESPONSE = "accounts_invalid_response"
    ACCOUNTS_LIST_EMPTY = "accounts_list_empty"
    ID_TOKEN_HTTP_NOT_FOUND = "id_token_http_not_found"
    ID_TOKEN_NO_RESPONSE = "id_token_no_response"
    ID_TOKEN_INVALID_RESPONSE = "id_token_invalid_response"
    ID_TOKEN_INVALID_REQUEST = "id_token_invalid_request"
    ERROR_ID_TOKEN = "error_id_token"
    CANCELED = "canceled"
    RP_PAGE_NOT_VISIBLE = "rp_page_not_visible"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class ClientHintIssueDetails:
    """This issue tracks client hints related issues.

    It's used to deprecate old features, encourage the use of new ones, and provide general guidance.
    """

    # Description is missing from the devtools protocol document. # noqa
    source_code_location: SourceCodeLocation
    # Description is missing from the devtools protocol document. # noqa
    client_hint_issue_reason: ClientHintIssueReason


class InspectorIssueCode(str, enum.Enum):
    """A unique identifier for the type of issue.

    Each type may use one of the optional fields in InspectorIssueDetails to convey more specific information about the
    kind of issue.
    """

    COOKIE_ISSUE = "cookie_issue"
    MIXED_CONTENT_ISSUE = "mixed_content_issue"
    BLOCKED_BY_RESPONSE_ISSUE = "blocked_by_response_issue"
    HEAVY_AD_ISSUE = "heavy_ad_issue"
    CONTENT_SECURITY_POLICY_ISSUE = "content_security_policy_issue"
    SHARED_ARRAY_BUFFER_ISSUE = "shared_array_buffer_issue"
    TRUSTED_WEB_ACTIVITY_ISSUE = "trusted_web_activity_issue"
    LOW_TEXT_CONTRAST_ISSUE = "low_text_contrast_issue"
    CORS_ISSUE = "cors_issue"
    ATTRIBUTION_REPORTING_ISSUE = "attribution_reporting_issue"
    QUIRKS_MODE_ISSUE = "quirks_mode_issue"
    NAVIGATOR_USER_AGENT_ISSUE = "navigator_user_agent_issue"
    GENERIC_ISSUE = "generic_issue"
    DEPRECATION_ISSUE = "deprecation_issue"
    CLIENT_HINT_ISSUE = "client_hint_issue"
    FEDERATED_AUTH_REQUEST_ISSUE = "federated_auth_request_issue"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class InspectorIssueDetails:
    """This struct holds a list of optional fields with additional information specific to the kind of issue.

    When adding a new issue code, please also add a new optional field to this type.
    """

    # Description is missing from the devtools protocol document. # noqa
    cookie_issue_details: typing.Optional[CookieIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    mixed_content_issue_details: typing.Optional[MixedContentIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    blocked_by_response_issue_details: typing.Optional[BlockedByResponseIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    heavy_ad_issue_details: typing.Optional[HeavyAdIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    content_security_policy_issue_details: typing.Optional[ContentSecurityPolicyIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    shared_array_buffer_issue_details: typing.Optional[SharedArrayBufferIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    twa_quality_enforcement_details: typing.Optional[TrustedWebActivityIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    low_text_contrast_issue_details: typing.Optional[LowTextContrastIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    cors_issue_details: typing.Optional[CorsIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    attribution_reporting_issue_details: typing.Optional[AttributionReportingIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    quirks_mode_issue_details: typing.Optional[QuirksModeIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    navigator_user_agent_issue_details: typing.Optional[NavigatorUserAgentIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    generic_issue_details: typing.Optional[GenericIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    deprecation_issue_details: typing.Optional[DeprecationIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    client_hint_issue_details: typing.Optional[ClientHintIssueDetails]
    # Description is missing from the devtools protocol document. # noqa
    federated_auth_request_issue_details: typing.Optional[FederatedAuthRequestIssueDetails]


class IssueId(str):
    """A unique id for a DevTools inspector issue.

    Allows other entities (e.g. exceptions, CDP message, console messages, etc.) to reference an issue.
    """

    def to_json(self) -> IssueId:
        return self

    @classmethod
    def from_json(cls, value: str) -> IssueId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


@dataclass
class InspectorIssue:
    """An inspector issue reported from the back-end."""

    # Description is missing from the devtools protocol document. # noqa
    code: InspectorIssueCode
    # Description is missing from the devtools protocol document. # noqa
    details: InspectorIssueDetails
    # A unique id for this issue. May be omitted if no other entity (e.g.exception, CDP message, etc.) is referencing this issue. # noqa
    issue_id: typing.Optional[IssueId]


@dataclass
@memoize_event("Audits.issueAdded")
class IssueAdded:
    """Description is missing from the devtools protocol document."""

    issue: InspectorIssue


async def get_encoded_response() -> None:
    """Returns the response body and size if it were re-encoded with the specified settings.

    Only applies to images. # noqa
    """
    ...


async def disable() -> None:
    """Disables issues domain, prevents further issues from being reported to the client.

    # noqa
    """
    ...


async def enable() -> None:
    """Enables issues domain, sends the issues collected so far to the client by means of the `issueAdded` event.

    # noqa
    """
    ...


async def check_contrast() -> None:
    """Runs the contrast check for the target page.

    Found issues are reported using Audits.issueAdded event. # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `Media`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/Media/

from __future__ import annotations

import typing
from dataclasses import dataclass

from .utils import memoize_event


class PlayerId(str):
    """Players will get an ID that is unique within the agent context."""

    def to_json(self) -> PlayerId:
        return self

    @classmethod
    def from_json(cls, value: str) -> PlayerId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class Timestamp(float):
    """Description is missing from the devtools protocol document."""

    def to_json(self) -> Timestamp:
        return self

    @classmethod
    def from_json(cls, value: float) -> Timestamp:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


@dataclass
class PlayerMessage:
    """Have one type per entry in MediaLogRecord::Type Corresponds to kMessage."""

    # Keep in sync with MediaLogMessageLevel We are currently keeping themessage level 'error' separate from the PlayerError type because right now theyrepresent different things, this one being a DVLOG(ERROR) style log message thatgets printed based on what log level is selected in the UI, and the other is arepresentation of a media::PipelineStatus object. Soon however we're going to bemoving away from using PipelineStatus for errors and introducing a new errortype which should hopefully let us integrate the error log level into thePlayerError type. # noqa
    level: typing.List[typing.Literal["error", "warning", "info", "debug"]]
    # Description is missing from the devtools protocol document. # noqa
    message: str


@dataclass
class PlayerProperty:
    """Corresponds to kMediaPropertyChange."""

    # Description is missing from the devtools protocol document. # noqa
    name: str
    # Description is missing from the devtools protocol document. # noqa
    value: str


@dataclass
class PlayerEvent:
    """Corresponds to kMediaEventTriggered."""

    # Description is missing from the devtools protocol document. # noqa
    timestamp: Timestamp
    # Description is missing from the devtools protocol document. # noqa
    value: str


@dataclass
class PlayerErrorSourceLocation:
    """Represents logged source line numbers reported in an error.

    NOTE: file and line are from chromium c++ implementation code, not js.
    """

    # Description is missing from the devtools protocol document. # noqa
    file: str
    # Description is missing from the devtools protocol document. # noqa
    line: int


@dataclass
class PlayerError:
    """Corresponds to kMediaError."""

    # Description is missing from the devtools protocol document. # noqa
    error_type: str
    # Code is the numeric enum entry for a specific set of error codes, such asPipelineStatusCodes in media/base/pipeline_status.h # noqa
    code: int
    # A trace of where this error was caused / where it passed through. # noqa
    stack: PlayerErrorSourceLocation
    # Errors potentially have a root cause error, ie, a DecoderError might becaused by an WindowsError # noqa
    cause: PlayerError
    # Extra data attached to an error, such as an HRESULT, Video Codec, etc. # noqa
    data: typing.Any


@dataclass
@memoize_event("Media.playerPropertiesChanged")
class PlayerPropertiesChanged:
    """This can be called multiple times, and can be used to set / override / remove player properties.

    A null propValue indicates removal.
    """

    player_id: PlayerId
    properties: typing.List[PlayerProperty]


@dataclass
@memoize_event("Media.playerEventsAdded")
class PlayerEventsAdded:
    """Send events as a list, allowing them to be batched on the browser for less congestion.

    If batched, events must ALWAYS be in chronological order.
    """

    player_id: PlayerId
    events: typing.List[PlayerEvent]


@dataclass
@memoize_event("Media.playerMessagesLogged")
class PlayerMessagesLogged:
    """Send a list of any messages that need to be delivered."""

    player_id: PlayerId
    messages: typing.List[PlayerMessage]


@dataclass
@memoize_event("Media.playerErrorsRaised")
class PlayerErrorsRaised:
    """Send a list of any errors that need to be delivered."""

    player_id: PlayerId
    errors: typing.List[PlayerError]


@dataclass
@memoize_event("Media.playersCreated")
class PlayersCreated:
    """Called whenever a player is created, or when a new agent joins and receives a list of active players.

    If an agent is restored, it will receive the full list of player ids and all events again.
    """

    players: typing.List[PlayerId]


async def enable() -> None:
    """Enables the Media domain # noqa."""
    ...


async def disable() -> None:
    """Disables the Media domain.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `WebAudio`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/WebAudio/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from .utils import memoize_event


class GraphObjectId(str):
    """An unique ID for a graph object (AudioContext, AudioNode, AudioParam) in Web Audio API."""

    def to_json(self) -> GraphObjectId:
        return self

    @classmethod
    def from_json(cls, value: str) -> GraphObjectId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class ContextType(str, enum.Enum):
    """Enum of BaseAudioContext types."""

    REALTIME = "realtime"
    OFFLINE = "offline"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class ContextState(str, enum.Enum):
    """Enum of AudioContextState from the spec."""

    SUSPENDED = "suspended"
    RUNNING = "running"
    CLOSED = "closed"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class NodeType(str):
    """Enum of AudioNode types."""

    def to_json(self) -> NodeType:
        return self

    @classmethod
    def from_json(cls, value: str) -> NodeType:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class ChannelCountMode(str, enum.Enum):
    """Enum of AudioNode::ChannelCountMode from the spec."""

    CLAMPED_MAX = "clamped-max"
    EXPLICIT = "explicit"
    MAX = "max"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class ChannelInterpretation(str, enum.Enum):
    """Enum of AudioNode::ChannelInterpretation from the spec."""

    DISCRETE = "discrete"
    SPEAKERS = "speakers"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class ParamType(str):
    """Enum of AudioParam types."""

    def to_json(self) -> ParamType:
        return self

    @classmethod
    def from_json(cls, value: str) -> ParamType:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class AutomationRate(str, enum.Enum):
    """Enum of AudioParam::AutomationRate from the spec."""

    A_RATE = "a-rate"
    K_RATE = "k-rate"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class ContextRealtimeData:
    """Fields in AudioContext that change in real-time."""

    # The current context time in second in BaseAudioContext. # noqa
    current_time: float
    # The time spent on rendering graph divided by render quantum duration, andmultiplied by 100. 100 means the audio renderer reached the full capacity andglitch may occur. # noqa
    render_capacity: float
    # A running mean of callback interval. # noqa
    callback_interval_mean: float
    # A running variance of callback interval. # noqa
    callback_interval_variance: float


@dataclass
class BaseAudioContext:
    """Protocol object for BaseAudioContext."""

    # Description is missing from the devtools protocol document. # noqa
    context_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    context_type: ContextType
    # Description is missing from the devtools protocol document. # noqa
    context_state: ContextState
    # Platform-dependent callback buffer size. # noqa
    callback_buffer_size: float
    # Number of output channels supported by audio hardware in use. # noqa
    max_output_channel_count: float
    # Context sample rate. # noqa
    sample_rate: float
    # Description is missing from the devtools protocol document. # noqa
    realtime_data: typing.Optional[ContextRealtimeData]


@dataclass
class AudioListener:
    """Protocol object for AudioListener."""

    # Description is missing from the devtools protocol document. # noqa
    listener_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    context_id: GraphObjectId


@dataclass
class AudioNode:
    """Protocol object for AudioNode."""

    # Description is missing from the devtools protocol document. # noqa
    node_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    context_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    node_type: NodeType
    # Description is missing from the devtools protocol document. # noqa
    number_of_inputs: float
    # Description is missing from the devtools protocol document. # noqa
    number_of_outputs: float
    # Description is missing from the devtools protocol document. # noqa
    channel_count: float
    # Description is missing from the devtools protocol document. # noqa
    channel_count_mode: ChannelCountMode
    # Description is missing from the devtools protocol document. # noqa
    channel_interpretation: ChannelInterpretation


@dataclass
class AudioParam:
    """Protocol object for AudioParam."""

    # Description is missing from the devtools protocol document. # noqa
    param_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    node_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    context_id: GraphObjectId
    # Description is missing from the devtools protocol document. # noqa
    param_type: ParamType
    # Description is missing from the devtools protocol document. # noqa
    rate: AutomationRate
    # Description is missing from the devtools protocol document. # noqa
    default_value: float
    # Description is missing from the devtools protocol document. # noqa
    min_value: float
    # Description is missing from the devtools protocol document. # noqa
    max_value: float


@dataclass
@memoize_event("WebAudio.contextCreated")
class ContextCreated:
    """Notifies that a new BaseAudioContext has been created."""

    context: BaseAudioContext


@dataclass
@memoize_event("WebAudio.contextWillBeDestroyed")
class ContextWillBeDestroyed:
    """Notifies that an existing BaseAudioContext will be destroyed."""

    context_id: GraphObjectId


@dataclass
@memoize_event("WebAudio.contextChanged")
class ContextChanged:
    """Notifies that existing BaseAudioContext has changed some properties (id stays the same).."""

    context: BaseAudioContext


@dataclass
@memoize_event("WebAudio.audioListenerCreated")
class AudioListenerCreated:
    """Notifies that the construction of an AudioListener has finished."""

    listener: AudioListener


@dataclass
@memoize_event("WebAudio.audioListenerWillBeDestroyed")
class AudioListenerWillBeDestroyed:
    """Notifies that a new AudioListener has been created."""

    context_id: GraphObjectId
    listener_id: GraphObjectId


@dataclass
@memoize_event("WebAudio.audioNodeCreated")
class AudioNodeCreated:
    """Notifies that a new AudioNode has been created."""

    node: AudioNode


@dataclass
@memoize_event("WebAudio.audioNodeWillBeDestroyed")
class AudioNodeWillBeDestroyed:
    """Notifies that an existing AudioNode has been destroyed."""

    context_id: GraphObjectId
    node_id: GraphObjectId


@dataclass
@memoize_event("WebAudio.audioParamCreated")
class AudioParamCreated:
    """Notifies that a new AudioParam has been created."""

    param: AudioParam


@dataclass
@memoize_event("WebAudio.audioParamWillBeDestroyed")
class AudioParamWillBeDestroyed:
    """Notifies that an existing AudioParam has been destroyed."""

    context_id: GraphObjectId
    node_id: GraphObjectId
    param_id: GraphObjectId


@dataclass
@memoize_event("WebAudio.nodesConnected")
class NodesConnected:
    """Notifies that two AudioNodes are connected."""

    context_id: GraphObjectId
    source_id: GraphObjectId
    destination_id: GraphObjectId
    source_output_index: typing.Optional[float]
    destination_input_index: typing.Optional[float]


@dataclass
@memoize_event("WebAudio.nodesDisconnected")
class NodesDisconnected:
    """Notifies that AudioNodes are disconnected.

    The destination can be null, and it means all the outgoing connections from the source are disconnected.
    """

    context_id: GraphObjectId
    source_id: GraphObjectId
    destination_id: GraphObjectId
    source_output_index: typing.Optional[float]
    destination_input_index: typing.Optional[float]


@dataclass
@memoize_event("WebAudio.nodeParamConnected")
class NodeParamConnected:
    """Notifies that an AudioNode is connected to an AudioParam."""

    context_id: GraphObjectId
    source_id: GraphObjectId
    destination_id: GraphObjectId
    source_output_index: typing.Optional[float]


@dataclass
@memoize_event("WebAudio.nodeParamDisconnected")
class NodeParamDisconnected:
    """Notifies that an AudioNode is disconnected to an AudioParam."""

    context_id: GraphObjectId
    source_id: GraphObjectId
    destination_id: GraphObjectId
    source_output_index: typing.Optional[float]


async def enable() -> None:
    """Enables the WebAudio domain and starts sending context lifetime events.

    # noqa
    """
    ...


async def disable() -> None:
    """Disables the WebAudio domain.

    # noqa
    """
    ...


async def get_realtime_data() -> None:
    """Fetch the realtime data from the registered contexts.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `LayerTree`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/LayerTree/

from __future__ import annotations

import typing
from dataclasses import dataclass

from . import dom
from .utils import memoize_event


class LayerId(str):
    """Unique Layer identifier."""

    def to_json(self) -> LayerId:
        return self

    @classmethod
    def from_json(cls, value: str) -> LayerId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class SnapshotId(str):
    """Unique snapshot identifier."""

    def to_json(self) -> SnapshotId:
        return self

    @classmethod
    def from_json(cls, value: str) -> SnapshotId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


@dataclass
class ScrollRect:
    """Rectangle where scrolling happens on the main thread."""

    # Rectangle itself. # noqa
    rect: dom.Rect
    # Reason for rectangle to force scrolling on the main thread # noqa
    type: typing.List[typing.Literal["RepaintsOnScroll", "TouchEventHandler", "WheelEventHandler"]]


@dataclass
class StickyPositionConstraint:
    """Sticky position constraints."""

    # Layout rectangle of the sticky element before being shifted # noqa
    sticky_box_rect: dom.Rect
    # Layout rectangle of the containing block of the sticky element # noqa
    containing_block_rect: dom.Rect
    # The nearest sticky layer that shifts the sticky box # noqa
    nearest_layer_shifting_sticky_box: typing.Optional[LayerId]
    # The nearest sticky layer that shifts the containing block # noqa
    nearest_layer_shifting_containing_block: typing.Optional[LayerId]


@dataclass
class PictureTile:
    """Serialized fragment of layer picture along with its offset within the layer."""

    # Offset from owning layer left boundary # noqa
    x: float
    # Offset from owning layer top boundary # noqa
    y: float
    # Base64-encoded snapshot data. (Encoded as a base64 string when passed overJSON) # noqa
    picture: str


@dataclass
class Layer:
    """Information about a compositing layer."""

    # The unique id for this layer. # noqa
    layer_id: LayerId
    # Offset from parent layer, X coordinate. # noqa
    offset_x: float
    # Offset from parent layer, Y coordinate. # noqa
    offset_y: float
    # Layer width. # noqa
    width: float
    # Layer height. # noqa
    height: float
    # Indicates how many time this layer has painted. # noqa
    paint_count: int
    # Indicates whether this layer hosts any content, rather than being used fortransform/scrolling purposes only. # noqa
    draws_content: bool
    # The id of parent (not present for root). # noqa
    parent_layer_id: typing.Optional[LayerId]
    # The backend id for the node associated with this layer. # noqa
    backend_node_id: typing.Optional[dom.BackendNodeId]
    # Transformation matrix for layer, default is identity matrix # noqa
    transform: typing.Optional[float]
    # Transform anchor point X, absent if no transform specified # noqa
    anchor_x: typing.Optional[float]
    # Transform anchor point Y, absent if no transform specified # noqa
    anchor_y: typing.Optional[float]
    # Transform anchor point Z, absent if no transform specified # noqa
    anchor_z: typing.Optional[float]
    # Set if layer is not visible. # noqa
    invisible: typing.Optional[bool]
    # Rectangles scrolling on main thread only. # noqa
    scroll_rects: typing.Optional[ScrollRect]
    # Sticky position constraint information # noqa
    sticky_position_constraint: typing.Optional[StickyPositionConstraint]


@dataclass
class PaintProfile:
    """Array of timings, one per paint step."""


@dataclass
@memoize_event("LayerTree.layerPainted")
class LayerPainted:
    """Description is missing from the devtools protocol document."""

    layer_id: LayerId
    clip: dom.Rect


@dataclass
@memoize_event("LayerTree.layerTreeDidChange")
class LayerTreeDidChange:
    """Description is missing from the devtools protocol document."""

    layers: typing.Optional[typing.List[Layer]]


async def compositing_reasons() -> None:
    """Provides the reasons why the given layer was composited.

    # noqa
    """
    ...


async def disable() -> None:
    """Disables compositing tree inspection.

    # noqa
    """
    ...


async def enable() -> None:
    """Enables compositing tree inspection.

    # noqa
    """
    ...


async def load_snapshot() -> None:
    """Returns the snapshot identifier.

    # noqa
    """
    ...


async def make_snapshot() -> None:
    """Returns the layer snapshot identifier.

    # noqa
    """
    ...


async def profile_snapshot() -> None:
    """Description is missing from the devtools protocol document.

    # noqa
    """
    ...


async def release_snapshot() -> None:
    """Releases layer snapshot captured by the back-end.

    # noqa
    """
    ...


async def replay_snapshot() -> None:
    """Replays the layer snapshot and returns the resulting bitmap.

    # noqa
    """
    ...


async def snapshot_command_log() -> None:
    """Replays the layer snapshot and returns canvas log.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `Tethering`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/Tethering/

from __future__ import annotations

from dataclasses import dataclass

from .utils import memoize_event


@dataclass
@memoize_event("Tethering.accepted")
class Accepted:
    """Informs that port was successfully bound and got a specified connection id."""

    port: int
    connection_id: str


async def bind() -> None:
    """Request browser port binding.

    # noqa
    """
    ...


async def unbind() -> None:
    """Request browser port unbinding.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `Tracing`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/Tracing/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from . import io
from .utils import memoize_event


@dataclass
class MemoryDumpConfig:
    """Configuration for memory dump.

    Used only when "memory-infra" category is enabled.
    """


@dataclass
class TraceConfig:
    """Description is missing from the devtools protocol document."""

    # Controls how the trace buffer stores data. # noqa
    record_mode: typing.Optional[
        typing.List[typing.Literal["recordUntilFull", "recordContinuously", "recordAsMuchAsPossible", "echoToConsole"]]
    ]
    # Size of the trace buffer in kilobytes. If not specified or zero is passed,a default value of 200 MB would be used. # noqa
    trace_buffer_size_in_kb: typing.Optional[float]
    # Turns on JavaScript stack sampling. # noqa
    enable_sampling: typing.Optional[bool]
    # Turns on system tracing. # noqa
    enable_systrace: typing.Optional[bool]
    # Turns on argument filter. # noqa
    enable_argument_filter: typing.Optional[bool]
    # Included category filters. # noqa
    included_categories: typing.Optional[str]
    # Excluded category filters. # noqa
    excluded_categories: typing.Optional[str]
    # Configuration to synthesize the delays in tracing. # noqa
    synthetic_delays: typing.Optional[str]
    # Configuration for memory dump triggers. Used only when "memory-infra"category is enabled. # noqa
    memory_dump_config: typing.Optional[MemoryDumpConfig]


class StreamFormat(str, enum.Enum):
    """Data format of a trace.

    Can be either the legacy JSON format or the protocol buffer format. Note that the JSON format will be deprecated
    soon.
    """

    JSON = "json"
    PROTO = "proto"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class StreamCompression(str, enum.Enum):
    """Compression type to use for traces returned via streams."""

    NONE = "none"
    GZIP = "gzip"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class MemoryDumpLevelOfDetail(str, enum.Enum):
    """Details exposed when memory request explicitly declared.

    Keep consistent with memory_dump_request_args.h and memory_instrumentation.mojom
    """

    BACKGROUND = "background"
    LIGHT = "light"
    DETAILED = "detailed"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class TracingBackend(str, enum.Enum):
    """Backend type to use for tracing.

    `chrome` uses the Chrome-integrated tracing service and is supported on all platforms. `system` is only supported on
    Chrome OS and uses the Perfetto system tracing service. `auto` chooses `system` when the perfettoConfig provided to
    Tracing.start specifies at least one non-Chrome data source; otherwise uses `chrome`.
    """

    AUTO = "auto"
    CHROME = "chrome"
    SYSTEM = "system"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
@memoize_event("Tracing.bufferUsage")
class BufferUsage:
    """Description is missing from the devtools protocol document."""

    percent_full: typing.Optional[float]
    event_count: typing.Optional[float]
    value: typing.Optional[float]


@dataclass
@memoize_event("Tracing.dataCollected")
class DataCollected:
    """Contains a bucket of collected trace events.

    When tracing is stopped collected events will be sent as a sequence of dataCollected events followed by
    tracingComplete event.
    """

    value: typing.List[typing.Any]


@dataclass
@memoize_event("Tracing.tracingComplete")
class TracingComplete:
    """Signals that tracing is stopped and there is no trace buffers pending flush, all data were delivered via
    dataCollected events."""

    data_loss_occurred: bool
    stream: typing.Optional[io.StreamHandle]
    trace_format: typing.Optional[StreamFormat]
    stream_compression: typing.Optional[StreamCompression]


async def end() -> None:
    """Stop trace events collection.

    # noqa
    """
    ...


async def get_categories() -> None:
    """Gets supported tracing categories.

    # noqa
    """
    ...


async def record_clock_sync_marker() -> None:
    """Record a clock sync marker in the trace.

    # noqa
    """
    ...


async def request_memory_dump() -> None:
    """Request a global memory dump.

    # noqa
    """
    ...


async def start() -> None:
    """Start trace events collection.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `DOMDebugger`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/DOMDebugger/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from . import dom
from . import runtime


class DOMBreakpointType(str, enum.Enum):
    """DOM breakpoint type."""

    SUBTREE_MODIFIED = "subtree-modified"
    ATTRIBUTE_MODIFIED = "attribute-modified"
    NODE_REMOVED = "node-removed"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class CSPViolationType(str, enum.Enum):
    """CSP Violation type."""

    TRUSTEDTYPE_SINK_VIOLATION = "trustedtype-sink-violation"
    TRUSTEDTYPE_POLICY_VIOLATION = "trustedtype-policy-violation"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class EventListener:
    """Object event listener."""

    # `EventListener`'s type. # noqa
    type: str
    # `EventListener`'s useCapture. # noqa
    use_capture: bool
    # `EventListener`'s passive flag. # noqa
    passive: bool
    # `EventListener`'s once flag. # noqa
    once: bool
    # Script id of the handler code. # noqa
    script_id: runtime.ScriptId
    # Line number in the script (0-based). # noqa
    line_number: int
    # Column number in the script (0-based). # noqa
    column_number: int
    # Event handler function value. # noqa
    handler: typing.Optional[runtime.RemoteObject]
    # Event original handler function value. # noqa
    original_handler: typing.Optional[runtime.RemoteObject]
    # Node the listener is added to (if any). # noqa
    backend_node_id: typing.Optional[dom.BackendNodeId]


async def get_event_listeners() -> None:
    """Returns event listeners of the given object.

    # noqa
    """
    ...


async def remove_dom_breakpoint() -> None:
    """Removes DOM breakpoint that was set using `setDOMBreakpoint`.

    # noqa
    """
    ...


async def remove_event_listener_breakpoint() -> None:
    """Removes breakpoint on particular DOM event.

    # noqa
    """
    ...


async def remove_instrumentation_breakpoint() -> None:
    """Removes breakpoint on particular native event.

    # noqa
    """
    ...


async def remove_xhr_breakpoint() -> None:
    """Removes breakpoint from XMLHttpRequest.

    # noqa
    """
    ...


async def set_break_on_csp_violation() -> None:
    """Sets breakpoint on particular CSP violations.

    # noqa
    """
    ...


async def set_dom_breakpoint() -> None:
    """Sets breakpoint on particular operation with DOM.

    # noqa
    """
    ...


async def set_event_listener_breakpoint() -> None:
    """Sets breakpoint on particular DOM event.

    # noqa
    """
    ...


async def set_instrumentation_breakpoint() -> None:
    """Sets breakpoint on particular native event.

    # noqa
    """
    ...


async def set_xhr_breakpoint() -> None:
    """Sets breakpoint on XMLHttpRequest.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `DOM`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/DOM/

from __future__ import annotations

import enum
import typing
from dataclasses import dataclass

from . import page
from .utils import memoize_event


class NodeId(int):
    """Unique DOM node identifier."""

    def to_json(self) -> NodeId:
        return self

    @classmethod
    def from_json(cls, value: int) -> NodeId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class BackendNodeId(int):
    """Unique DOM node identifier used to reference a node that may not have been pushed to the front-end."""

    def to_json(self) -> BackendNodeId:
        return self

    @classmethod
    def from_json(cls, value: int) -> BackendNodeId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


@dataclass
class BackendNode:
    """Backend node with a friendly name."""

    # `Node`'s nodeType. # noqa
    node_type: int
    # `Node`'s nodeName. # noqa
    node_name: str
    # Description is missing from the devtools protocol document. # noqa
    backend_node_id: BackendNodeId


class PseudoType(str, enum.Enum):
    """Pseudo element type."""

    FIRST_LINE = "first-line"
    FIRST_LETTER = "first-letter"
    BEFORE = "before"
    AFTER = "after"
    MARKER = "marker"
    BACKDROP = "backdrop"
    SELECTION = "selection"
    TARGET_TEXT = "target-text"
    SPELLING_ERROR = "spelling-error"
    GRAMMAR_ERROR = "grammar-error"
    HIGHLIGHT = "highlight"
    FIRST_LINE_INHERITED = "first-line-inherited"
    SCROLLBAR = "scrollbar"
    SCROLLBAR_THUMB = "scrollbar-thumb"
    SCROLLBAR_BUTTON = "scrollbar-button"
    SCROLLBAR_TRACK = "scrollbar-track"
    SCROLLBAR_TRACK_PIECE = "scrollbar-track-piece"
    SCROLLBAR_CORNER = "scrollbar-corner"
    RESIZER = "resizer"
    INPUT_LIST_BUTTON = "input-list-button"
    VIEW_TRANSITION = "view-transition"
    VIEW_TRANSITION_GROUP = "view-transition-group"
    VIEW_TRANSITION_IMAGE_PAIR = "view-transition-image-pair"
    VIEW_TRANSITION_OLD = "view-transition-old"
    VIEW_TRANSITION_NEW = "view-transition-new"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class ShadowRootType(str, enum.Enum):
    """Shadow root type."""

    USER_AGENT = "user-agent"
    OPEN = "open"
    CLOSED = "closed"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class CompatibilityMode(str, enum.Enum):
    """Document compatibility mode."""

    QUIRKS_MODE = "quirks_mode"
    LIMITED_QUIRKS_MODE = "limited_quirks_mode"
    NO_QUIRKS_MODE = "no_quirks_mode"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class PhysicalAxes(str, enum.Enum):
    """ContainerSelector physical axes."""

    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"
    BOTH = "both"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


class LogicalAxes(str, enum.Enum):
    """ContainerSelector logical axes."""

    INLINE = "inline"
    BLOCK = "block"
    BOTH = "both"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class Node:
    """DOM interaction is implemented in terms of mirror objects that represent the actual DOM nodes.

    DOMNode is a base node mirror type.
    """

    # Node identifier that is passed into the rest of the DOM messages as the`nodeId`. Backend will only push node with given `id` once. It is aware of allrequested nodes and will only fire DOM events for nodes known to the client. # noqa
    node_id: NodeId
    # The BackendNodeId for this node. # noqa
    backend_node_id: BackendNodeId
    # `Node`'s nodeType. # noqa
    node_type: int
    # `Node`'s nodeName. # noqa
    node_name: str
    # `Node`'s localName. # noqa
    local_name: str
    # `Node`'s nodeValue. # noqa
    node_value: str
    # The id of the parent node if any. # noqa
    parent_id: typing.Optional[NodeId]
    # Child count for `Container` nodes. # noqa
    child_node_count: typing.Optional[int]
    # Child nodes of this node when requested with children. # noqa
    children: typing.Optional[Node]
    # Attributes of the `Element` node in the form of flat array `[name1,value1, name2, value2]`. # noqa
    attributes: typing.Optional[str]
    # Document URL that `Document` or `FrameOwner` node points to. # noqa
    document_url: typing.Optional[str]
    # Base URL that `Document` or `FrameOwner` node uses for URL completion. # noqa
    base_url: typing.Optional[str]
    # `DocumentType`'s publicId. # noqa
    public_id: typing.Optional[str]
    # `DocumentType`'s systemId. # noqa
    system_id: typing.Optional[str]
    # `DocumentType`'s internalSubset. # noqa
    internal_subset: typing.Optional[str]
    # `Document`'s XML version in case of XML documents. # noqa
    xml_version: typing.Optional[str]
    # `Attr`'s name. # noqa
    name: typing.Optional[str]
    # `Attr`'s value. # noqa
    value: typing.Optional[str]
    # Pseudo element type for this node. # noqa
    pseudo_type: typing.Optional[PseudoType]
    # Pseudo element identifier for this node. Only present if there is a validpseudoType. # noqa
    pseudo_identifier: typing.Optional[str]
    # Shadow root type. # noqa
    shadow_root_type: typing.Optional[ShadowRootType]
    # Frame ID for frame owner elements. # noqa
    frame_id: typing.Optional[page.FrameId]
    # Content document for frame owner elements. # noqa
    content_document: typing.Optional[Node]
    # Shadow root list for given element host. # noqa
    shadow_roots: typing.Optional[Node]
    # Content document fragment for template elements. # noqa
    template_content: typing.Optional[Node]
    # Pseudo elements associated with this node. # noqa
    pseudo_elements: typing.Optional[Node]
    # Deprecated, as the HTML Imports API has been removed (crbug.com/937746).This property used to return the imported document for the HTMLImport links. Theproperty is always undefined now. # noqa
    imported_document: typing.Optional[Node]
    # Distributed nodes for given insertion point. # noqa
    distributed_nodes: typing.Optional[BackendNode]
    # Whether the node is SVG. # noqa
    is_svg: typing.Optional[bool]
    # Description is missing from the devtools protocol document. # noqa
    compatibility_mode: typing.Optional[CompatibilityMode]
    # Description is missing from the devtools protocol document. # noqa
    assigned_slot: typing.Optional[BackendNode]


@dataclass
class RGBA:
    """A structure holding an RGBA color."""

    # The red component, in the [0-255] range. # noqa
    r: int
    # The green component, in the [0-255] range. # noqa
    g: int
    # The blue component, in the [0-255] range. # noqa
    b: int
    # The alpha component, in the [0-1] range (default: 1). # noqa
    a: typing.Optional[float]


@dataclass
class Quad:
    """An array of quad vertices, x immediately followed by y for each point, points clock-wise."""


@dataclass
class BoxModel:
    """Box model."""

    # Content box # noqa
    content: Quad
    # Padding box # noqa
    padding: Quad
    # Border box # noqa
    border: Quad
    # Margin box # noqa
    margin: Quad
    # Node width # noqa
    width: int
    # Node height # noqa
    height: int
    # Shape outside coordinates # noqa
    shape_outside: typing.Optional[ShapeOutsideInfo]


@dataclass
class ShapeOutsideInfo:
    """CSS Shape Outside details."""

    # Shape bounds # noqa
    bounds: Quad
    # Shape coordinate details # noqa
    shape: typing.Any
    # Margin shape bounds # noqa
    margin_shape: typing.Any


@dataclass
class Rect:
    """Rectangle."""

    # X coordinate # noqa
    x: float
    # Y coordinate # noqa
    y: float
    # Rectangle width # noqa
    width: float
    # Rectangle height # noqa
    height: float


@dataclass
class CSSComputedStyleProperty:
    """Description is missing from the devtools protocol document."""

    # Computed style property name. # noqa
    name: str
    # Computed style property value. # noqa
    value: str


@dataclass
@memoize_event("DOM.attributeModified")
class AttributeModified:
    """Fired when `Element`'s attribute is modified."""

    node_id: NodeId
    name: str
    value: str


@dataclass
@memoize_event("DOM.attributeRemoved")
class AttributeRemoved:
    """Fired when `Element`'s attribute is removed."""

    node_id: NodeId
    name: str


@dataclass
@memoize_event("DOM.characterDataModified")
class CharacterDataModified:
    """Mirrors `DOMCharacterDataModified` event."""

    node_id: NodeId
    character_data: str


@dataclass
@memoize_event("DOM.childNodeCountUpdated")
class ChildNodeCountUpdated:
    """Fired when `Container`'s child node count has changed."""

    node_id: NodeId
    child_node_count: int


@dataclass
@memoize_event("DOM.childNodeInserted")
class ChildNodeInserted:
    """Mirrors `DOMNodeInserted` event."""

    parent_node_id: NodeId
    previous_node_id: NodeId
    node: Node


@dataclass
@memoize_event("DOM.childNodeRemoved")
class ChildNodeRemoved:
    """Mirrors `DOMNodeRemoved` event."""

    parent_node_id: NodeId
    node_id: NodeId


@dataclass
@memoize_event("DOM.distributedNodesUpdated")
class DistributedNodesUpdated:
    """Called when distribution is changed."""

    insertion_point_id: NodeId
    distributed_nodes: typing.List[BackendNode]


@dataclass
@memoize_event("DOM.documentUpdated")
class DocumentUpdated:
    """Fired when `Document` has been totally updated.

    Node ids are no longer valid.
    """


@dataclass
@memoize_event("DOM.inlineStyleInvalidated")
class InlineStyleInvalidated:
    """Fired when `Element`'s inline style is modified via a CSS property modification."""

    node_ids: typing.List[NodeId]


@dataclass
@memoize_event("DOM.pseudoElementAdded")
class PseudoElementAdded:
    """Called when a pseudo element is added to an element."""

    parent_id: NodeId
    pseudo_element: Node


@dataclass
@memoize_event("DOM.topLayerElementsUpdated")
class TopLayerElementsUpdated:
    """Called when top layer elements are changed."""


@dataclass
@memoize_event("DOM.pseudoElementRemoved")
class PseudoElementRemoved:
    """Called when a pseudo element is removed from an element."""

    parent_id: NodeId
    pseudo_element_id: NodeId


@dataclass
@memoize_event("DOM.setChildNodes")
class SetChildNodes:
    """Fired when backend wants to provide client with the missing DOM structure.

    This happens upon most of the calls requesting node ids.
    """

    parent_id: NodeId
    nodes: typing.List[Node]


@dataclass
@memoize_event("DOM.shadowRootPopped")
class ShadowRootPopped:
    """Called when shadow root is popped from the element."""

    host_id: NodeId
    root_id: NodeId


@dataclass
@memoize_event("DOM.shadowRootPushed")
class ShadowRootPushed:
    """Called when shadow root is pushed into the element."""

    host_id: NodeId
    root: Node


async def collect_class_names_from_subtree() -> None:
    """Collects class names for the node with given id and all of it's child nodes.

    # noqa
    """
    ...


async def copy_to() -> None:
    """Creates a deep copy of the specified node and places it into the target container before the given anchor.

    # noqa
    """
    ...


async def describe_node() -> None:
    """Describes node given its id, does not require domain to be enabled.

    Does not start tracking any objects, can be used for automation. # noqa
    """
    ...


async def scroll_into_view_if_needed() -> None:
    """Scrolls the specified rect of the given node into view if not already visible.

    Note: exactly one between nodeId, backendNodeId and objectId should be passed
    to identify the node. # noqa
    """
    ...


async def disable() -> None:
    """Disables DOM agent for the given page.

    # noqa
    """
    ...


async def discard_search_results() -> None:
    """Discards search results from the session with the given id.

    `getSearchResults` should no longer be called for that search. # noqa
    """
    ...


async def enable() -> None:
    """Enables DOM agent for the given page.

    # noqa
    """
    ...


async def focus() -> None:
    """Focuses the given element.

    # noqa
    """
    ...


async def get_attributes() -> None:
    """Returns attributes for the specified node.

    # noqa
    """
    ...


async def get_box_model() -> None:
    """Returns boxes for the given node.

    # noqa
    """
    ...


async def get_content_quads() -> None:
    """Returns quads that describe node position on the page.

    This method might return multiple quads for inline nodes. # noqa
    """
    ...


async def get_document() -> None:
    """Returns the root DOM node (and optionally the subtree) to the caller.

    Implicitly enables the DOM domain events for the current target. # noqa
    """
    ...


async def get_flattened_document() -> None:
    """Returns the root DOM node (and optionally the subtree) to the caller.

    Deprecated, as it is not designed to work well with the rest of the DOM agent. Use DOMSnapshot.captureSnapshot
    instead. # noqa
    """
    ...


async def get_nodes_for_subtree_by_style() -> None:
    """Finds nodes with a given computed style in a subtree.

    # noqa
    """
    ...


async def get_node_for_location() -> None:
    """Returns node id at given location.

    Depending on whether DOM domain is enabled, nodeId is either returned or not. # noqa
    """
    ...


async def get_outer_html() -> None:
    """Returns node's HTML markup.

    # noqa
    """
    ...


async def get_relayout_boundary() -> None:
    """Returns the id of the nearest ancestor that is a relayout boundary.

    # noqa
    """
    ...


async def get_search_results() -> None:
    """Returns search results from given `fromIndex` to given `toIndex` from the search with the given identifier.

    # noqa
    """
    ...


async def hide_highlight() -> None:
    """Hides any highlight.

    # noqa
    """
    ...


async def highlight_node() -> None:
    """Highlights DOM node.

    # noqa
    """
    ...


async def highlight_rect() -> None:
    """Highlights given rectangle.

    # noqa
    """
    ...


async def mark_undoable_state() -> None:
    """Marks last undoable state.

    # noqa
    """
    ...


async def move_to() -> None:
    """Moves node into the new container, places it before the given anchor.

    # noqa
    """
    ...


async def perform_search() -> None:
    """Searches for a given string in the DOM tree.

    Use `getSearchResults` to access search results or `cancelSearch` to end this search session. # noqa
    """
    ...


async def push_node_by_path_to_frontend() -> None:
    """Requests that the node is sent to the caller given its path.

    // FIXME, use XPath # noqa
    """
    ...


async def push_nodes_by_backend_ids_to_frontend() -> None:
    """Requests that a batch of nodes is sent to the caller given their backend node ids.

    # noqa
    """
    ...


async def query_selector() -> None:
    """Executes `querySelector` on a given node.

    # noqa
    """
    ...


async def query_selector_all() -> None:
    """Executes `querySelectorAll` on a given node.

    # noqa
    """
    ...


async def get_top_layer_elements() -> None:
    """Returns NodeIds of current top layer elements.

    Top layer is rendered closest to the user within a viewport, therefore its elements always appear on top of all
    other content. # noqa
    """
    ...


async def redo() -> None:
    """Re-does the last undone action.

    # noqa
    """
    ...


async def remove_attribute() -> None:
    """Removes attribute with given name from an element with given id.

    # noqa
    """
    ...


async def remove_node() -> None:
    """Removes node with given id.

    # noqa
    """
    ...


async def request_child_nodes() -> None:
    """Requests that children of the node with given id are returned to the caller in form of `setChildNodes` events
    where not only immediate children are retrieved, but all children down to the specified depth.

    # noqa
    """
    ...


async def request_node() -> None:
    """Requests that the node is sent to the caller given the JavaScript node object reference.

    All nodes that form the path from the node to the root are also sent to the client as a series of `setChildNodes`
    notifications. # noqa
    """
    ...


async def resolve_node() -> None:
    """Resolves the JavaScript node object for a given NodeId or BackendNodeId.

    # noqa
    """
    ...


async def set_attribute_value() -> None:
    """Sets attribute for an element with given id.

    # noqa
    """
    ...


async def set_attributes_as_text() -> None:
    """Sets attributes on element with given id.

    This method is useful when user edits some existing attribute value and types in several attribute name/value pairs.
    # noqa
    """
    ...


async def set_file_input_files() -> None:
    """Sets files for the given file input element.

    # noqa
    """
    ...


async def set_node_stack_traces_enabled() -> None:
    """Sets if stack traces should be captured for Nodes.

    See `Node.getNodeStackTraces`. Default is disabled. # noqa
    """
    ...


async def get_node_stack_traces() -> None:
    """Gets stack traces associated with a Node.

    As of now, only provides stack trace for Node creation. # noqa
    """
    ...


async def get_file_info() -> None:
    """Returns file information for the given File wrapper.

    # noqa
    """
    ...


async def set_inspected_node() -> None:
    """Enables console to refer to the node with given id via $x (see Command Line API for more details.

    $x functions). # noqa
    """
    ...


async def set_node_name() -> None:
    """Sets node name for a node with given id.

    # noqa
    """
    ...


async def set_node_value() -> None:
    """Sets node value for a node with given id.

    # noqa
    """
    ...


async def set_outer_html() -> None:
    """Sets node HTML markup, returns new node id.

    # noqa
    """
    ...


async def undo() -> None:
    """Undoes the last performed action.

    # noqa
    """
    ...


async def get_frame_owner() -> None:
    """Returns iframe node that owns iframe with the given domain.

    # noqa
    """
    ...


async def get_container_for_node() -> None:
    """Returns the query container of the given node based on container query
    conditions: containerName, physical, and logical axes. If no axes are
    provided, the style container is returned, which is the direct parent or the
    closest element with a matching container-name. # noqa"""
    ...


async def get_querying_descendants_for_container() -> None:
    """Returns the descendants of a container query container that have container queries against this container.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `BackgroundService`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/

from __future__ import annotations

import enum
from dataclasses import dataclass

from . import network
from . import service_worker
from .utils import memoize_event


class ServiceName(str, enum.Enum):
    """The Background Service that will be associated with the commands/events.

    Every Background Service operates independently, but they share the same API.
    """

    BACKGROUND_FETCH = "background_fetch"
    BACKGROUND_SYNC = "background_sync"
    PUSH_MESSAGING = "push_messaging"
    NOTIFICATIONS = "notifications"
    PAYMENT_HANDLER = "payment_handler"
    PERIODIC_BACKGROUND_SYNC = "periodic_background_sync"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class EventMetadata:
    """A key-value pair for additional event information to pass along."""

    # Description is missing from the devtools protocol document. # noqa
    key: str
    # Description is missing from the devtools protocol document. # noqa
    value: str


@dataclass
class BackgroundServiceEvent:
    """Description is missing from the devtools protocol document."""

    # Timestamp of the event (in seconds). # noqa
    timestamp: network.TimeSinceEpoch
    # The origin this event belongs to. # noqa
    origin: str
    # The Service Worker ID that initiated the event. # noqa
    service_worker_registration_id: service_worker.RegistrationID
    # The Background Service this event belongs to. # noqa
    service: ServiceName
    # A description of the event. # noqa
    event_name: str
    # An identifier that groups related events together. # noqa
    instance_id: str
    # A list of event-specific information. # noqa
    event_metadata: EventMetadata
    # Storage key this event belongs to. # noqa
    storage_key: str


@dataclass
@memoize_event("BackgroundService.recordingStateChanged")
class RecordingStateChanged:
    """Called when the recording state for the service has been updated."""

    is_recording: bool
    service: ServiceName


@dataclass
@memoize_event("BackgroundService.backgroundServiceEventReceived")
class BackgroundServiceEventReceived:
    """Called with all existing backgroundServiceEvents when enabled, and all new events afterwards if enabled and
    recording."""

    background_service_event: BackgroundServiceEvent


async def start_observing() -> None:
    """Enables event updates for the service.

    # noqa
    """
    ...


async def stop_observing() -> None:
    """Disables event updates for the service.

    # noqa
    """
    ...


async def set_recording() -> None:
    """Set the recording state for the service.

    # noqa
    """
    ...


async def clear_events() -> None:
    """Clears all stored data for the service.

    # noqa
    """
    ...



import functools
import json
import pathlib
import re
import textwrap
import typing

import stringcase


def indent(text: str, by: int = 4) -> str:
    """Utility method for indenting text, using spaces.

    :param text: The text to indent.
    :param by: How many spaces to use.
    """
    return textwrap.indent(text=text, prefix=" " * by)


def resolve_docstring(docs: str) -> str:
    """Handles edge cases with docstring management and returns the docstring in a suitable format."""
    return indent(f'""" {docs} """\n')


@functools.cache
def api_type_to_python_annotation(api_type: str) -> str:
    """Returns the pythonic type for a type return by the CDP api."""
    return {
        "integer": "int",
        "string": "str",
        "number": "float",
        "boolean": "bool",
        "object": "typing.Any",
    }.get(
        api_type,
        api_type,
    )


def parse_javascript_specification() -> typing.Dict[str, typing.Any]:
    """Parse the javascript protocol specification."""
    with open(str(get_specification_rootdir() / "js_protocol.json")) as f:
        return json.load(f)


def parse_browser_specification() -> typing.Dict[str, typing.Any]:
    """Parse the browser protocol specification."""
    with open(str(get_specification_rootdir() / "browser_protocol.json")) as f:
        return json.load(f)


def get_specification_rootdir() -> pathlib.Path:
    """Returns the directory containing the specification files from the submodule."""
    return pathlib.Path(__file__).parents[1] / "devtools-protocol" / "json"


def get_generation_rootdir() -> pathlib.Path:
    """Returns the rootdirectory to the CDP generated root dir."""
    return pathlib.Path(__file__).parent / "cdp"


def name_to_snake_case(name: str):
    """Given a string; convert it to camel case.

    Taken from https://stackoverflow.com/a/1176023
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def name_to_pascal_case(name: str) -> str:
    """Converts any name to a pythonic class name."""
    return stringcase.pascalcase(name)


def camel_to_enum_member(name: str) -> str:
    """Converts a name to a pythonic enum member.

    This is words seperated by `_` all in upper case.
    """
    word = ""
    for idx, char in enumerate(name.replace("-", "_")):
        if char.isupper() and idx != 0:
            word += f"_{char}"
            continue
        word += char.upper()
    return word



from __future__ import annotations  # isort: skip
import collections
import itertools
import pathlib
import textwrap
import typing
from dataclasses import dataclass

from ._const import MISSING_DESCRIPTION_IN_PROTOCOL_DOC
from ._headers import CONSTANT_IMPORTS
from ._headers import PREAMBLE
from ._protocols import GeneratesSourceCode
from ._protocols import Requirable
from ._templates import PRIMITIVE_TYPE_FROM_JSON
from ._templates import PRIMITIVE_TYPE_TO_JSON
from ._templates import SIMPLE_ENUM_FROM_JSON
from ._templates import SIMPLE_PRIMITIVE_REPR
from ._types import AnyDict
from ._utils import api_type_to_python_annotation
from ._utils import camel_to_enum_member
from ._utils import get_generation_rootdir
from ._utils import indent
from ._utils import name_to_pascal_case
from ._utils import name_to_snake_case
from ._utils import resolve_docstring

TypeStore = collections.namedtuple("TypeStore", "parent, annotation")
PRIMITIVE_TYPE_FACTORY = {
    "string": TypeStore("str", "str"),
    "number": TypeStore("float", "float"),
    "boolean": TypeStore("bool", "bool"),
    "integer": TypeStore("int", "int"),
}


def fix_reference_type_circular(domain: str, reference: str) -> str:
    """Patches a small bug in the protocol where certain properties or parameters are advertised as reliant on a type
    that lives in the same module once we have split files by top level `domains`.  The bug is outlined below:

    Occassionally some files i.e `dom.py` have a reference to a type defined within
    that module but the devtools protocol returns data such as `dom.Type`.  Due to
    how we handle imports using `from .domain import X` this is not valid and so
    we manage this internally.  I have opened a defect on the CDP tracker but it was
    marked P3 and unlikely to be resolved soon.

    :param domain: The domain name to check for a circle reference defect.
    :param reference: The reference returned by CDP.
    """
    if reference == "any":
        return "typing.Any"
    if "." in reference:
        split_domain, period, annotation = reference.partition(".")
        if domain.lower() == split_domain.lower():
            return annotation  # Defect, referring to a type in the same module.
        return f"{name_to_snake_case(split_domain)}{period}{annotation}"
    return reference


@dataclass
class DevtoolsArrayItem:
    """Encapsulation of a property `item` array entry."""

    type: typing.Optional[str] = None
    ref: typing.Optional[str] = None

    @classmethod
    def from_json(cls, payload: AnyDict) -> DevtoolsArrayItem:
        return cls(type=payload.get("type"), ref=payload.get("$ref"))


@dataclass
class DevtoolsParam:
    """Encapsulation of a parameter passed to a method call or event type."""

    cdp_domain: str
    name: str
    description: str
    optional: bool
    ref: typing.Optional[str] = None
    type: typing.Optional[str] = None
    items: typing.Optional[DevtoolsArrayItem] = None
    enum_options: typing.Optional[typing.List[str]] = None

    @classmethod
    def from_json(cls, payload: AnyDict, cdp_domain: str) -> DevtoolsParam:
        return cls(
            cdp_domain=cdp_domain,
            name=typing.cast(str, payload.get("name")),
            description=payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC),
            optional=payload.get("optional", False),
            ref=payload.get("$ref"),
            type=payload.get("type"),
            items=DevtoolsArrayItem.from_json(payload["items"]) if "items" in payload else None,
            enum_options=payload.get("enum"),
        )

    def generate_code(self) -> str:
        source = ""
        source += indent(f"{name_to_snake_case(self.name)}: {self.generate_type_hint()}")
        source += "\n"
        return source

    def generate_type_hint(self) -> str:
        """Generates the type hint for this parameter."""
        optional = "typing.Optional[{}]" if self.optional else "{}"
        if self.items is not None:
            if self.items.ref:
                return optional.format(f"typing.List[{self.items.ref}]")
            return optional.format(f"typing.List[{api_type_to_python_annotation(self.items.type)}]")
        if self.enum_options:
            # An array of primitive or reference types let's pythonise the options tho.
            wrapped = ", ".join(f"'{name_to_snake_case(word)}'" for word in self.enum_options)
            return optional.format(f"typing.Literal[{wrapped}]")
        if self.type:
            # various types, could even be an object, read from mapping
            return optional.format(api_type_to_python_annotation(self.type))
        if self.ref:
            # object reference type, return it literally
            return optional.format(fix_reference_type_circular(self.cdp_domain, self.ref))
        return "typing.Any"

    @property
    def requires(self) -> typing.Set[str]:
        """Return the required imports for this particular parameter."""
        imports = set()
        reference = (self.items.ref or self.items.type) if self.items else self.ref
        if reference:
            fixed = fix_reference_type_circular(self.cdp_domain, reference)
            if "." in fixed:
                fixed = fixed.split(".")[0]
                imports.add(f"from . import {fixed}")
        return imports


@dataclass
class DevtoolsReturn:
    """Encapsulation of a method return value."""

    name: str
    description: str
    optional: typing.Optional[bool] = None
    type: typing.Optional[str] = None
    items: typing.Optional[DevtoolsArrayItem] = None
    ref: typing.Optional[str] = None

    @classmethod
    def from_json(cls, payload: AnyDict) -> DevtoolsReturn:
        return cls(
            name=typing.cast(str, payload.get("name")),
            description=payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC),
            optional=payload.get("optional", False),
            type=payload.get("type"),
            items=DevtoolsArrayItem.from_json(payload["items"]) if "items" in payload else None,
            ref=payload.get("ref"),
        )


@dataclass
class DevtoolsProperty:
    """Encapsulation of a property for objects that are not simple primitive types."""

    cdp_domain: str
    name: str
    description: str
    items: typing.Optional[DevtoolsArrayItem] = None
    ref: typing.Optional[str] = None
    optional: typing.Optional[bool] = False
    type: typing.Optional[str] = None
    enum_options: typing.Optional[typing.List[str]] = None

    @classmethod
    def from_json(cls, payload: AnyDict, cdp_domain: str) -> DevtoolsProperty:
        return cls(
            cdp_domain=cdp_domain,
            name=typing.cast(str, payload.get("name")),
            description=payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC),
            ref=payload.get("$ref", None),
            optional=payload.get("optional", False),
            items=DevtoolsArrayItem.from_json(payload["items"]) if "items" in payload else None,
            type=payload.get("type", None),
            enum_options=payload.get("enum"),
        )

    def generate_code(self) -> str:
        """Generate the source code for this particular type property."""
        source = ""
        source += "".join(textwrap.wrap(self.description, width=80, initial_indent="    # "))
        source += " # noqa"  # Todo: Remove this and wrap appropriately.
        source += "\n"
        source += self.generate_definition()
        source += "\n"
        return source

    def generate_definition(self) -> str:
        optional = "typing.Optional[{}]" if self.optional else "{}"
        literal = "typing.Literal[{}]" or "{}"
        seq = "typing.List[{}]" if (self.enum_options or self.type == "array") else "{}"
        source = indent(f"{name_to_snake_case(self.name)}: ")
        if self.enum_options:
            return source + optional.format(
                seq.format(literal.format(", ".join(f"'{word}'" for word in self.enum_options))),
            )
        if self.type == "array" and self.items:
            reference = self.items.ref or self.items.type
            assert reference is not None, "item array for property but both are none!"
            if reference in PRIMITIVE_TYPE_FACTORY:
                return source + optional.format(api_type_to_python_annotation(reference))
            return source + optional.format(fix_reference_type_circular(self.cdp_domain, reference))
        if self.type in PRIMITIVE_TYPE_FACTORY:
            return source + optional.format(api_type_to_python_annotation(self.type))
        if self.ref:
            return source + optional.format(fix_reference_type_circular(self.cdp_domain, self.ref))
        return source + "typing.Any"

    @property
    def requires(self) -> typing.Set[str]:
        imports = set()
        reference = (self.items.ref or self.items.type) if self.items else self.ref
        if reference and "." in reference:
            domain, _, _ = reference.partition(".")
            imports.add(f"from . import {name_to_snake_case(domain)}")
        return imports


@dataclass
class DevtoolsType:
    cdp_domain: str
    id: str
    description: str
    type: str
    properties: typing.List[DevtoolsProperty]
    enum_options: typing.List[str]

    @classmethod
    def from_json(cls, payload: AnyDict, cdp_domain: str) -> DevtoolsType:
        return cls(
            cdp_domain=cdp_domain,
            id=typing.cast(str, payload.get("id")),
            description=payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC),
            type=typing.cast(str, payload.get("type")),
            # These need kept in order with non optional fields first to avoid pain with the generated dataclasses.
            properties=list(
                sorted(
                    [DevtoolsProperty.from_json(p, cdp_domain=cdp_domain) for p in payload.get("properties", [])],
                    key=lambda i: i.optional,  # type: ignore [arg-type, return-value]
                ),
            ),
            enum_options=payload.get("enum", []),
        )

    def generate_code(self) -> str:
        """Build out python source code for the CDP types."""
        if self.type in PRIMITIVE_TYPE_FACTORY and not self.enum_options:
            # A simple primitive type wrapper will suffice for this class
            return self._build_for_primitive_type()
        elif self.enum_options:
            # A string enum type wrapper is necessary
            return self._build_for_enum_type()
        return self._build_for_object_type()

    def _build_for_enum_type(self) -> str:
        """Generate source code for enum types."""
        source = textwrap.dedent(f"\n\nclass {self.id}(str, enum.Enum):")
        source += "\n"
        source += resolve_docstring(self.description)
        source += "\n"
        for option in self.enum_options:
            source += indent(f'{camel_to_enum_member(option)} = "{name_to_snake_case(option)}"')
            source += "\n"
        source += "\n"
        source += indent(SIMPLE_ENUM_FROM_JSON)
        return source

    def _build_for_object_type(self) -> str:
        """Generate source code for object types."""
        source = "\n"
        source += f'''
@dataclass
class {self.id}:
    """ {self.description} """
'''
        for property in self.properties:
            source += property.generate_code()
        return source

    def _build_for_primitive_type(self) -> str:
        """Generate source code for primitive types (simple subclass wrappers)."""
        source = "\n\n"
        source += textwrap.dedent(f"class {self.id}({PRIMITIVE_TYPE_FACTORY[self.type].parent}):")
        source += "\n"
        source += resolve_docstring(self.description)
        source += indent(PRIMITIVE_TYPE_TO_JSON.format(self.id))
        source += "\n"
        source += indent(PRIMITIVE_TYPE_FROM_JSON.format(PRIMITIVE_TYPE_FACTORY[self.type].parent, self.id))
        source += "\n"
        source += indent(SIMPLE_PRIMITIVE_REPR.format("{self.__class__.__name__}", "({super().__repr__()})"))
        return source

    @property
    def requires(self) -> typing.Set[str]:
        imports = set()
        for property in self.properties:
            imports |= property.requires
        return imports


@dataclass
class DevtoolsEvent:
    """Encapsulation of a CDP Event that can be sent and received across the websocket/connection."""

    cdp_domain: str
    name: str
    description: str
    parameters: typing.Optional[typing.List[DevtoolsParam]]
    experimental: typing.Optional[bool]
    deprecated: typing.Optional[bool]

    def generate_code(self) -> str:
        """Generate the source for every event object advertised in the protocol."""
        source = textwrap.dedent("@dataclass")
        source += "\n"
        source += textwrap.dedent(f"@memoize_event('{self.cdp_domain}.{self.name}')")
        source += "\n"
        source += f"class {self.class_name}:"
        source += "\n"
        source += resolve_docstring(self.description)
        if self.parameters:
            for param in self.parameters:
                source += param.generate_code()
        return source

    @property
    def class_name(self) -> str:
        """Generates the pythonic class name for this event."""
        return name_to_pascal_case(self.name)

    @property
    def requires(self) -> typing.Set[str]:
        """Returns a distinct set of the imports required based on the parameters."""
        imports: typing.Set[str] = set()
        if self.parameters is None:
            return imports
        imports.add("from .utils import memoize_event")
        for parameter in self.parameters:
            imports |= parameter.requires
        return imports

    @classmethod
    def from_json(cls, payload: AnyDict, cdp_domain: str) -> DevtoolsEvent:
        """Generate the event object, including building its nested parameters."""
        return cls(
            cdp_domain=cdp_domain,
            name=typing.cast(str, payload["name"]),
            description=typing.cast(str, payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC)),
            parameters=[DevtoolsParam.from_json(p, cdp_domain=cdp_domain) for p in payload.get("parameters", [])],
            experimental=typing.cast(bool, payload.get("experimental", False)),
            deprecated=typing.cast(bool, payload.get("deprecated", False)),
        )


@dataclass
class DevtoolsCommand:
    """Encapsulation of a devtools command."""

    cdp_domain: str
    name: str
    description: str
    experimental: typing.Optional[bool]
    parameters: typing.Optional[typing.List[DevtoolsParam]]
    returns: typing.Optional[typing.List[DevtoolsReturn]]

    def generate_code(self) -> str:
        source = textwrap.dedent(
            f'''
async def {name_to_snake_case(self.name)}() -> None:
    """ {self.description} # noqa """
    ...
''',
        )
        return source

    @classmethod
    def from_json(cls, payload: AnyDict, cdp_domain: str) -> DevtoolsCommand:
        return cls(
            cdp_domain=cdp_domain,
            name=typing.cast(str, payload.get("name")),
            description=typing.cast(str, payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC)),
            experimental=payload.get("experimental", False),
            parameters=[DevtoolsParam.from_json(p, cdp_domain) for p in payload.get("parameters", [])],
            returns=[DevtoolsReturn.from_json(r) for r in payload.get("returns", [])],
        )


@dataclass
class DevtoolsDomain:
    """Encapsulation of an individual devtools domain."""

    domain: str
    description: str
    dependencies: typing.List[str]
    deprecated: bool
    experimental: bool
    events: typing.List[DevtoolsEvent]
    types: typing.List[DevtoolsType]
    commands: typing.List[DevtoolsCommand]

    @property
    def py_mod_name(self) -> str:
        """Returns the python module name for this instance."""
        return f"{name_to_snake_case(self.domain)}.py"

    @classmethod
    def from_json(cls, payload: AnyDict) -> DevtoolsDomain:
        """Shovel the json arguments into this model and recursively build out nested objects."""
        return cls(
            domain=typing.cast(str, payload.get("domain")),
            description=payload.get("description", MISSING_DESCRIPTION_IN_PROTOCOL_DOC),
            deprecated=payload.get("deprecated", False),
            dependencies=payload.get("dependencies", []),
            experimental=payload.get("experimental", False),
            events=[DevtoolsEvent.from_json(e, cdp_domain=payload["domain"]) for e in payload.get("events", [])],
            types=[DevtoolsType.from_json(t, cdp_domain=payload["domain"]) for t in payload.get("types", [])],
            commands=[
                DevtoolsCommand.from_json(c, typing.cast(str, payload.get("domain")))
                for c in payload.get("commands", [])
            ],
        )

    def generate_code(self) -> str:
        """Generate the full source code for the domain module."""
        requires_enum = bool([x for x in self.types if x.enum_options])  # Todo: investigate logic.
        source = PREAMBLE.format(domain=self.domain)
        source += "\n"
        source += CONSTANT_IMPORTS
        if requires_enum:
            source += textwrap.dedent("import enum")
        source += "\n\n"
        source += self.calculate_imports()
        iterator: typing.Iterator[GeneratesSourceCode] = itertools.chain(self.types, self.events, self.commands)
        for item in iterator:
            source += "\n\n"
            source += item.generate_code()
        return source

    def calculate_imports(self) -> str:
        """Calculate the correct imports for this module based on builtins and other domains created in the cdp/*
        directory.

        There might be some shortcomings here in the protocol itself,
        raise bugs/fixes for those respectively.

        Referencing the `dependencies` array is likely not sufficient
        here.  We need
        to calculate them dynamically (better) and sort potential edge
        cases and
        race conditions with order etc.

        This has a heap of bugs already
            ** Audit has no dependency on `Runtime` in the protocol.
            ** Audit has no dependency on `DOM` in the protocol.
            ** Audit has no dependency on `Page` in the protocol.

            ** BackgroundService has no dependency on `Network` in the protocol.
            ** BackgroundService has no dependency on `ServiceWorker` in the protocol.

            ** Dom has no dependency on `Page` in the protocol.
            ** Network depends on itself in error
        """
        # self.commands add later, using raw `dependencies` on the `domain` is also not the best solution for now.
        elements: typing.Iterable[Requirable] = itertools.chain(self.events, self.types)
        imports = set()
        for element in elements:
            imports |= element.requires

        # Todo: Format inline with linter, or leave that up to the linter?
        source = ""
        source += "\n".join(imports)
        # Todo: Handle joining them
        # Todo: Rewrite the docstring here.
        return source

    def create_py_module(self) -> pathlib.Path:
        """Writes the python module for this domain to disk, recursively generating all the python source code."""
        path = get_generation_rootdir() / self.py_mod_name
        with open(str(path), mode="w") as f:
            f.write(self.generate_code())
        return path


@dataclass
class TopLevelDomains:
    """Encapsulation of the top level domains array.  This is composed of an array of DevtoolDomain objects.

    For now, only domains in the protocol that are not marked `deprecated` are built and accessible, but this is likely
    subject to change in future.
    """

    domains: typing.List[DevtoolsDomain]

    def __iter__(self) -> typing.Iterator[DevtoolsDomain]:
        return iter(self.domains)

    @classmethod
    def from_json(cls, payload: AnyDict) -> TopLevelDomains:
        """Build and generate the full protocol."""
        return cls(domains=[DevtoolsDomain.from_json(domain) for domain in payload["domains"]])

    def create_source_code_on_disk(self) -> None:
        """Automatically generates all the python CDP modules for all of the nested children domains."""
        for domain in self.domains:
            domain.create_py_module()



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `Schema`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/Schema/

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Domain:
    """Description of the protocol domain."""

    # Domain name. # noqa
    name: str
    # Domain version. # noqa
    version: str


async def get_domains() -> None:
    """Returns supported domains.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `HeadlessExperimental`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/HeadlessExperimental/

from __future__ import annotations

import typing
from dataclasses import dataclass


@dataclass
class ScreenshotParams:
    """Encoding options for a screenshot."""

    # Image compression format (defaults to png). # noqa
    format: typing.Optional[typing.List[typing.Literal["jpeg", "png", "webp"]]]
    # Compression quality from range [0..100] (jpeg only). # noqa
    quality: typing.Optional[int]
    # Optimize image encoding for speed, not for resulting size (defaults tofalse) # noqa
    optimize_for_speed: typing.Optional[bool]


async def begin_frame() -> None:
    """Sends a BeginFrame to the target and returns when the frame was completed.

    Optionally captures a screenshot from the resulting frame. Requires that the target was created with enabled
    BeginFrameControl. Designed for use with --run-all-compositor-stages-before-draw, see also
    https://goo.gle/chrome-headless-rendering
    for more background. # noqa
    """
    ...


async def disable() -> None:
    """Disables headless events for the target.

    # noqa
    """
    ...


async def enable() -> None:
    """Enables headless events for the target.

    # noqa
    """
    ...



# THIS FILE HAS BEEN AUTOMATICALLY GENERATED.
# CHANGES TO THIS FILE ARE FUTILE AS IT WILL BE OVERWRITTEN
# WHEN THE DEVTOOLS PROTOCOL CHANGES.  IF THERE ANY BUGS
# OR YOU WISH TO CHANGE HOW THE FILES ARE GENERATED PLEASE
# REFER TO: https://github.com/symonk/python-cdp OR YOUR
# OWN FORK.  REFERENCE THE `generate.py` FILE FOR CONTEXT
# AND INSTRUCTIONS.
# Chrome Devtools Protocol Domain Mapped to: `CacheStorage`.
# Url for domain: https://chromedevtools.github.io/devtools-protocol/tot/CacheStorage/

from __future__ import annotations

import enum
from dataclasses import dataclass


class CacheId(str):
    """Unique identifier of the Cache object."""

    def to_json(self) -> CacheId:
        return self

    @classmethod
    def from_json(cls, value: str) -> CacheId:
        return cls(value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(({super().__repr__()}))"


class CachedResponseType(str, enum.Enum):
    """Type of HTTP response cached."""

    BASIC = "basic"
    CORS = "cors"
    DEFAULT = "default"
    ERROR = "error"
    OPAQUE_RESPONSE = "opaque_response"
    OPAQUE_REDIRECT = "opaque_redirect"

    @classmethod
    def from_json(cls, value: str) -> str:
        return cls(value)


@dataclass
class DataEntry:
    """Data entry."""

    # Request URL. # noqa
    request_url: str
    # Request method. # noqa
    request_method: str
    # Request headers # noqa
    request_headers: Header
    # Number of seconds since epoch. # noqa
    response_time: float
    # HTTP response status code. # noqa
    response_status: int
    # HTTP response status text. # noqa
    response_status_text: str
    # HTTP response type # noqa
    response_type: CachedResponseType
    # Response headers # noqa
    response_headers: Header


@dataclass
class Cache:
    """Cache identifier."""

    # An opaque unique id of the cache. # noqa
    cache_id: CacheId
    # Security origin of the cache. # noqa
    security_origin: str
    # Storage key of the cache. # noqa
    storage_key: str
    # The name of the cache. # noqa
    cache_name: str


@dataclass
class Header:
    """Description is missing from the devtools protocol document."""

    # Description is missing from the devtools protocol document. # noqa
    name: str
    # Description is missing from the devtools protocol document. # noqa
    value: str


@dataclass
class CachedResponse:
    """Cached response."""

    # Entry content, base64-encoded. (Encoded as a base64 string when passedover JSON) # noqa
    body: str


async def delete_cache() -> None:
    """Deletes a cache.

    # noqa
    """
    ...


async def delete_entry() -> None:
    """Deletes a cache entry.

    # noqa
    """
    ...


async def request_cache_names() -> None:
    """Requests cache names.

    # noqa
    """
    ...


async def request_cached_response() -> None:
    """Fetches cache entry.

    # noqa
    """
    ...


async def request_entries() -> None:
    """Requests data from cache.

    # noqa
    """
    ...
